"""The Metador widget server."""
import io
from typing import Dict, List, Literal, Optional, Union

import numpy as np
from bokeh.application import Application
from bokeh.application.handlers.function import FunctionHandler
from bokeh.document import Document
from bokeh.embed import server_document
from bokeh.server.server import Server
from flask import Blueprint, request, send_file
from panel.viewable import Viewable
from tornado.ioloop import IOLoop
from werkzeug.exceptions import BadRequest, NotFound

from metador_core.container import ContainerProxy, MetadorContainer, MetadorNode


class WidgetServer:
    """Server backing the instances of Metador widgets (and dashboard).

    Metador widgets depend on a `WidgetServer` to:
    * get data from Metador containers (via special flask API, provided as a mountable blueprint)
    * wire up the information flow with a bokeh server instance (requirement for interactive bokeh widgets)

    For information on running a bokeh server see:
    https://docs.bokeh.org/en/latest/docs/user_guide/server.html#embedding-bokeh-server-as-a-library
    """

    @classmethod
    def _get_widget_arg(cls, args: Dict[str, List[bytes]], name: str) -> Optional[str]:
        """Extract argument from bokeh server request argument dict."""
        return args[name][0].decode("utf-8") if name in args and args[name] else None

    @classmethod
    def _get_widget_args(cls, doc: Document):
        """Extract arguments from bokeh server request parameters."""
        args = doc.session_context.request.arguments
        return dict(
            container_id=cls._get_widget_arg(args, "id"),
            container_path=cls._get_widget_arg(args, "path"),
        )

    @classmethod
    def _make_widget_args(
        cls, container_id: str, container_path: Optional[str]
    ) -> Dict[str, str]:
        """Construct dict to be passed through bokeh request into widget."""
        req_args = {"id": container_id}
        if container_path:
            req_args["path"] = container_path
        return req_args

    def _get_bokeh_widget_name(
        self,
        viewable_type: Literal["widget", "dashboard"],
        name: str,
    ) -> str:
        """Return mapped name of a registered widget or dashboard (bokeh server endpoint).

        Raises NotFound exception if widget has not been found.
        """
        if viewable_type not in {"widget", "dashboard"}:
            msg = f"Invalid type: {viewable_type}. Must be widget or dashboard!"
            raise NotFound(msg)
        known = self._reg_widgets if viewable_type == "widget" else self._reg_dashboards
        if name not in known:
            raise NotFound(f"Bokeh {viewable_type} not found: '{name}'")
        return known[name]

    def _get_container_node(
        self, container_id: str, container_path: Optional[str] = None
    ) -> Optional[Union[MetadorContainer, MetadorNode]]:
        """Retrieve desired container (and target path, if provided).

        If `path` is provided in the query parameters,
        will return the container node, otherwise returns the full container.

        Raises NotFound exception if container or path in container do not exist.
        """
        try:
            container = self._containers.get(container_id)
        except KeyError as e:
            raise NotFound(f"Container not found: '{container_id}'") from e

        if container_path is None:
            return container
        if node := container.get(container_path):
            return node.restrict(read_only=True, local_only=True)

        raise NotFound(f"Path not found in container: {container_path}")

    # ----

    def __init__(
        self,
        containers: ContainerProxy[str],
        *,
        bokeh_endpoint: Optional[str] = None,
        flask_endpoint: Optional[str] = None,
        populate: bool = True,
    ):
        """Widget server to serve widget- and dashboard-like bokeh entities.

        Args:
            containers: `ContainerProxy` to retrieve containers by some container id string.
            bokeh_endpoint: Endpoint where the bokeh server will run (`WidgetServer.run()`)
            flask_endpoint: Endpoint where Widget API is mounted (`WidgetServer.get_flask_blueprint()`)
            populate: If true (default), load and serve all installed widgets and generic dashboard
        """
        self._containers = containers
        self._bokeh_apps: Dict[str, Application] = {}
        self._reg_widgets: Dict[str, str] = {}
        self._reg_dashboards: Dict[str, str] = {}

        # these can be set after launching the server threads
        # (e.g. in case of dynamic port selection)
        self._flask_endpoint = flask_endpoint or ""
        self._bokeh_endpoint = bokeh_endpoint or ""

        if populate:
            self.register_installed()

    def register_installed(self) -> None:
        """Register installed widgets and the generic dashboard."""
        # NOTE: do imports here, otherwise circular imports.
        from metador_core.plugins import widgets

        from ..dashboard import Dashboard

        self.register_dashboard("generic", self.make_bokeh_app(Dashboard))
        for wclass in widgets.values():
            self.register_widget(
                wclass.Plugin.plugin_string(), self.make_bokeh_app(wclass)
            )

    def register_widget(self, name: str, bokeh_app: Application) -> None:
        """Register a new widget application."""
        mapped_name = f"w-{name}"
        self._bokeh_apps[f"/{mapped_name}"] = bokeh_app
        self._reg_widgets[name] = mapped_name

    def register_dashboard(self, name: str, bokeh_app: Application) -> None:
        """Register a new dashboard application."""
        mapped_name = f"d-{name}"
        self._bokeh_apps[f"/{mapped_name}"] = bokeh_app
        self._reg_dashboards[name] = mapped_name

    def make_bokeh_app(self, viewable_class: Viewable) -> Application:
        def handler(doc: Document) -> None:
            """Return bokeh app for Metador widget.

            In this context, a suitable class must satisfy the interface
            of being initialized with a metador node or container,
            and having a `show()` method returning a panel `Viewable`.

            The app will understand take `id` and optionally a `path` as query params.
            These are parsed and used to look up the correct container (node).
            """
            w_args = self._get_widget_args(doc)
            if c_obj := self._get_container_node(**w_args):
                # if we retrieved container / node, instantiate a widget and show it
                widget = viewable_class(
                    c_obj, server=self, container_id=w_args["container_id"]
                ).show()
                doc.add_root(widget.get_root(doc))

        return Application(FunctionHandler(handler, trap_exceptions=True))

    @property
    def flask_endpoint(self) -> str:
        """Get configured endpoint where WidgetServer API is mounted."""
        return self._flask_endpoint

    @flask_endpoint.setter
    def flask_endpoint(self, uri: str):
        """Set URI where the blueprint from `get_flask_blueprint` is mounted."""
        self._flask_endpoint = uri.rstrip("/")

    @property
    def bokeh_endpoint(self) -> str:
        """Get URI where the bokeh server is running."""
        return self._bokeh_endpoint

    @bokeh_endpoint.setter
    def bokeh_endpoint(self, uri: str):
        """Set URI where the bokeh server is running."""
        self._bokeh_endpoint = uri.rstrip("/")

    def run(self, **kwargs):
        """Run bokeh server with the registered apps (will block the current process)."""
        # kwargs["io_loop"] = kwargs.get("io_loop") or IOLoop()
        # server = pn.io.server.get_server(self._bokeh_apps, **kwargs)

        kwargs["loop"] = kwargs.get("io_loop") or IOLoop()
        server = Server(self._bokeh_apps, **kwargs)

        server.start()
        server.io_loop.start()

    # ----
    # Helper functions exposed to widgets

    def file_url_for(self, container_id: str, node: MetadorNode) -> str:
        """Return URL for given container ID and file at Metador Container node.

        To be used by widgets that need direct access to files in the container.
        """
        if not self._flask_endpoint:
            raise RuntimeError("missing flask endpoint!")
        return f"{self._flask_endpoint}/file/{container_id}{node.name}"

    # ----
    # Functions making up the WidgetServer API

    def index(self):
        """Return information about current Metador environment.

        Response includes an overview of metador-related Python packages,
        Metador plugins, and the known widgets (nodes) and dashboards (containers).
        """
        from metador_core.plugin.types import to_ep_name
        from metador_core.plugins import plugingroups

        # build dict with all available metador plugins
        pgs = {to_ep_name(x.name, x.version): x.dict() for x in plugingroups.keys()}
        groups = {plugingroups.Plugin.name: pgs}
        for pg in plugingroups.values():
            groups[pg.Plugin.name] = {
                to_ep_name(x.name, x.version): x.dict() for x in pg.keys()
            }

        return {
            "widgets": list(self._reg_widgets),
            "dashboards": list(self._reg_dashboards),
            "plugins": groups,
        }

    def download(self, container_id: str, container_path: str):
        """Return file download stream of a file embedded in the container."""
        node = self._get_container_node(container_id, container_path)
        # get data out of container
        obj = node[()]
        bs = obj.tolist() if isinstance(obj, np.void) else obj
        if not isinstance(bs, bytes):
            raise BadRequest(f"Path not a bytes object: /{container_path}")

        # construct a default file name based on path in container
        def_name = f"{container_id}_{container_path.replace('/', '__')}"
        # if object has attached file metadata, use it to serve data:
        filemeta = node.meta.get("core.file")
        name = filemeta.id_ if filemeta else def_name
        mime = filemeta.encodingFormat if filemeta else None

        # requested as explicit file download?
        dl = bool(request.args.get("download", False))
        # return file download stream with download metadata
        return send_file(
            io.BytesIO(bs), download_name=name, mimetype=mime, as_attachment=dl
        )

    def get_script(
        self,
        viewable_type: Literal["widget", "dashboard"],
        name: str,
        container_id: str,
        container_path: Optional[str] = None,
    ) -> str:
        """Return a script tag that will auto-load the desired widget for selected container."""
        if not self._bokeh_endpoint:
            raise RuntimeError("missing bokeh endpoint!")
        if viewable_type == "dashboard" and container_path:
            raise BadRequest("Dashboards do not accept a container path!")

        return server_document(
            f"{self._bokeh_endpoint}/{self._get_bokeh_widget_name(viewable_type, name)}",
            arguments=self._make_widget_args(container_id, container_path),
        )

    def get_flask_blueprint(self, *args):
        """Return a Flask blueprint with the Metador container and widget API."""
        api = Blueprint(*args)

        api.route("/")(self.index)
        api.route("/file/<container_id>/<path:container_path>")(self.download)
        api.route("/<viewable_type>/<name>/<container_id>/")(
            api.route("/<viewable_type>/<name>/<container_id>/<path:container_path>")(
                self.get_script
            )
        )

        return api


# NOTE: snippet to make script tag not evaluate by default
# it can be used to prevent the auto-loading during DOM injection, if needed for some reason
#     # disable self-evaluation (save call in variable, call when requested)
#     sc_id = re.match(r"\s*<script id=\"(.*)\">", script).group(1)
#     script = script.replace("(function", f"w{sc_id} = (function").replace(
#         "})();", "});"
#     )
#     return f'{script}<button type="button" onclick="w{sc_id}()">Load widget</button>'



from __future__ import annotations

import json
from dataclasses import dataclass
from enum import Enum, auto
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    ItemsView,
    Iterator,
    KeysView,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    ValuesView,
    cast,
    overload,
)
from uuid import UUID, uuid1

from typing_extensions import TypeAlias

from ..plugin.types import EPName, from_ep_name, plugin_args, to_ep_name
from ..plugins import schemas
from ..schema import MetadataSchema
from ..schema.plugins import PluginPkgMeta, PluginRef
from ..schema.types import SemVerTuple
from . import utils as M
from .drivers import (
    METADOR_DRIVERS,
    MetadorDriver,
    MetadorDriverEnum,
    get_driver_type,
    get_source,
)
from .protocols import H5DatasetLike, H5FileLike, H5GroupLike

if TYPE_CHECKING:
    from .wrappers import MetadorContainer, MetadorNode

S = TypeVar("S", bound=MetadataSchema)


class NodeAcl(Enum):
    """Metador node soft access control flags.

    Soft means - they can be bypassed, it is about trying to prevent errors.

    Group nodes inherit their ACL flags to child nodes.
    """

    # NOTE: maybe refactor this to IntFlag? Then e.g. restrict() interface can be like:
    # node.restrict(acl=NodeAcl.read_only | NodeAcl.local_only)

    read_only = auto()
    """Forbid calling methods mutating contents of (meta)data."""

    local_only = auto()
    """Forbid access to parents beyond the initial local node."""

    skel_only = auto()
    """Forbid reading datasets and metadata, only existence can be checked."""


NodeAclFlags: TypeAlias = Dict[NodeAcl, bool]


@dataclass
class StoredMetadata:
    """Information about a metadata schema instance stored at a node."""

    uuid: UUID
    """UUID identifying the metadata object in the container.

    Used for bookkeeping, i.e. keeping the container TOC in sync.
    """

    schema: PluginRef
    """Schema the object is an instance of."""

    node: H5DatasetLike
    """Node with serialized metadata object."""

    def to_path(self):
        """Return path of metadata object.

        (E.g. to return canonical path for copying TOC link nodes)
        """
        prefix = self.node.parent.name
        ep_name = to_ep_name(self.schema.name, self.schema.version)
        return f"{prefix}/{ep_name}={self.uuid}"

    @staticmethod
    def from_node(obj: H5DatasetLike) -> StoredMetadata:
        """Instantiate info about a stored metadata node."""
        path = obj.name
        segs = path.lstrip("/").split("/")
        ep_name, uuid_str = segs.pop().split("=")
        s_name, s_vers = from_ep_name(EPName(ep_name))
        uuid = UUID(uuid_str)
        s_ref = schemas.PluginRef(name=s_name, version=s_vers)
        return StoredMetadata(uuid=uuid, schema=s_ref, node=obj)


def _schema_ref_for(ep_name: str) -> PluginRef:
    s_name, s_ver = from_ep_name(EPName(ep_name))
    return schemas.PluginRef(name=s_name, version=s_ver)


def _ep_name_for(s_ref: PluginRef) -> str:
    return to_ep_name(s_ref.name, s_ref.version)


class MetadorMeta:
    """Interface to Metador metadata objects stored at a single HDF5 node."""

    # helpers for __getitem__ and __setitem__

    @staticmethod
    def _require_schema(
        schema_name: str, schema_ver: Optional[SemVerTuple]
    ) -> Type[MetadataSchema]:
        """Return compatible installed schema class, if possible.

        Raises KeyError if no suitable schema was found.

        Raises TypeError if an auxiliary schema is requested.
        """
        schema_class = schemas._get_unsafe(
            schema_name, schema_ver
        )  # can raise KeyError
        if schema_class.Plugin.auxiliary:  # reject auxiliary schemas in container
            msg = f"Cannot attach instances of auxiliary schema '{schema_name}' to a node!"
            raise TypeError(msg)
        return schema_class

    @staticmethod
    def _parse_obj(
        schema: Type[S], obj: Union[str, bytes, Dict[str, Any], MetadataSchema]
    ) -> S:
        """Return original object if it is an instance of passed schema, or else parse it.

        Raises ValidationError if parsing fails.
        """
        if isinstance(obj, schema):
            return obj  # skip validation, already correct model!
        # try to convert/parse it:
        if isinstance(obj, (str, bytes)):
            return schema.parse_raw(obj)
        if isinstance(obj, MetadataSchema):
            return schema.parse_obj(obj.dict())
        else:  # dict
            return schema.parse_obj(obj)

    # raw getters and setters don't care about the environment,
    # they work only based on what objects are available and compatible
    # and do not perform validation etc.

    def _get_raw(
        self, schema_name: str, version: Optional[SemVerTuple] = None
    ) -> Optional[StoredMetadata]:
        """Return stored metadata for given schema at this node (or None).

        If a version is passed, the stored version must also be compatible.
        """
        # retrieve stored instance (if suitable)
        ret: Optional[StoredMetadata] = self._objs.get(schema_name)
        if not version:
            return ret  # no specified version -> anything goes
        # otherwise: only return if it is compatible
        req_ref: Optional[PluginRef] = None
        req_ref = schemas.PluginRef(name=schema_name, version=version)
        return ret if ret and req_ref.supports(ret.schema) else None

    def _set_raw(self, schema_ref: PluginRef, obj: MetadataSchema) -> None:
        """Store metadata object as instance of passed schema at this node."""
        # reserve UUID, construct dataset path and store metadata object
        obj_uuid = self._mc.metador._links.fresh_uuid()
        obj_path = f"{self._base_dir}/{_ep_name_for(schema_ref)}={str(obj_uuid)}"
        # store object
        self._mc.__wrapped__[obj_path] = bytes(obj)
        obj_node = self._mc.__wrapped__[obj_path]
        assert isinstance(obj_node, H5DatasetLike)
        stored_obj = StoredMetadata(uuid=obj_uuid, schema=schema_ref, node=obj_node)
        self._objs[schema_ref] = stored_obj
        # update TOC
        self._mc.metador._links.register(stored_obj)
        return

    def _del_raw(self, schema_name: str, *, _unlink: bool = True) -> None:
        """Delete stored metadata for given schema at this node."""
        # NOTE: _unlink is only for the destroy method
        stored_obj = self._objs[schema_name]
        # unregister in TOC (will also trigger clean up there)
        if _unlink:
            self._mc.metador._links.unregister(stored_obj.uuid)
        # remove metadata object
        del self._objs[stored_obj.schema.name]
        del self._mc.__wrapped__[stored_obj.node.name]
        # no metadata objects left -> remove metadata dir
        if not self._objs:
            del self._mc.__wrapped__[self._base_dir]
        return

    # helpers for container-level opertions (move, copy, delete etc)

    def _destroy(self, *, _unlink: bool = True):
        """Unregister and delete all metadata objects attached to this node."""
        # NOTE: _unlink is only set to false for node copy without metadata
        for schema_name in list(self.keys()):
            self._del_raw(schema_name, _unlink=_unlink)

    # ----

    def __init__(self, node: MetadorNode):
        self._mc: MetadorContainer = node._self_container
        """Underlying container (for convenience)."""

        self._node: MetadorNode = node
        """Underlying actual user node."""

        is_dataset = isinstance(node, H5DatasetLike)
        self._base_dir: str = M.to_meta_base_path(node.name, is_dataset)
        """Path of this metador metadata group node.

        Actual node exists iff any metadata is stored for the node.
        """

        self._objs: Dict[str, StoredMetadata] = {}
        """Information about available metadata objects."""

        # load available object metadata encoded in the node names
        meta_grp = cast(H5GroupLike, self._mc.__wrapped__.get(self._base_dir, {}))
        for obj_node in meta_grp.values():
            assert isinstance(obj_node, H5DatasetLike)
            obj = StoredMetadata.from_node(obj_node)
            self._objs[obj.schema.name] = obj

    # ----

    def keys(self) -> KeysView[str]:
        """Return names of explicitly attached metadata objects.

        Transitive parent schemas are not included.
        """
        return self._objs.keys()

    def values(self) -> ValuesView[StoredMetadata]:
        self._node._guard_acl(NodeAcl.skel_only)
        return self._objs.values()

    def items(self) -> ItemsView[str, StoredMetadata]:
        self._node._guard_acl(NodeAcl.skel_only)
        return self._objs.items()

    # ----

    def __len__(self) -> int:
        """Return number of explicitly attached metadata objects.

        Transitive parent schemas are not counted.
        """
        return len(self.keys())

    def __iter__(self) -> Iterator[str]:
        """Iterate listing schema names of all actually attached metadata objects.

        Transitive parent schemas are not included.
        """
        return iter(self.keys())

    # ----

    def query(
        self,
        schema: Union[
            str, Tuple[str, Optional[SemVerTuple]], PluginRef, Type[MetadataSchema]
        ] = "",
        version: Optional[SemVerTuple] = None,
    ) -> Iterator[PluginRef]:
        """Return schema names for which objects at this node are compatible with passed schema.

        Will also consider compatible child schema instances.

        Returned iterator will yield passed schema first, if an object is available.
        Apart from this, the order is not specified.
        """
        schema_name, schema_ver = plugin_args(schema, version)
        # no schema selected -> list everything
        if not schema_name:
            for obj in self.values():
                yield obj.schema
            return

        # try exact schema (in any compatible version, if version specified)
        if obj := self._get_raw(schema_name, schema_ver):
            yield obj.schema

        # next, try compatible child schemas of compatible versions of requested schema
        compat = set().union(
            *(
                self._mc.metador.schemas.children(ref)
                for ref in self._mc.metador.schemas.versions(schema_name, schema_ver)
            )
        )
        avail = {self._get_raw(s).schema for s in self.keys()}
        for s_ref in avail.intersection(compat):
            yield s_ref

    def __contains__(
        self,
        schema: Union[
            str, Tuple[str, Optional[SemVerTuple]], PluginRef, Type[MetadataSchema]
        ],
    ) -> bool:
        """Check whether a compatible metadata object for given schema exists.

        Will also consider compatible child schema instances.
        """
        if schema == "" or isinstance(schema, tuple) and schema[0] == "":
            return False  # empty query lists everything, here the logic is inverted!
        return next(self.query(schema), None) is not None

    @overload
    def __getitem__(self, schema: str) -> MetadataSchema:
        ...

    @overload
    def __getitem__(self, schema: Type[S]) -> S:
        ...

    def __getitem__(self, schema: Union[str, Type[S]]) -> Union[S, MetadataSchema]:
        """Like get, but will raise KeyError on failure."""
        if ret := self.get(schema):
            return ret
        raise KeyError(schema)

    @overload
    def get(
        self, schema: str, version: Optional[SemVerTuple] = None
    ) -> Optional[MetadataSchema]:
        ...

    @overload
    def get(
        self, schema: Type[S], version: Optional[SemVerTuple] = None
    ) -> Optional[S]:
        ...

    def get(
        self, schema: Union[str, Type[S]], version: Optional[SemVerTuple] = None
    ) -> Optional[Union[MetadataSchema, S]]:
        """Get a parsed metadata object matching the given schema (if it exists).

        Will also consider compatible child schema instances.
        """
        self._node._guard_acl(NodeAcl.skel_only)

        # normalize arguments
        schema_name, schema_ver = plugin_args(schema, version)

        # get a compatible schema instance that is available at this node
        compat_schema = next(self.query(schema_name, schema_ver), None)
        if not compat_schema:
            return None  # not found

        # get class of schema and parse object
        schema_class = self._require_schema(schema_name, schema_ver)
        if obj := self._get_raw(compat_schema.name, compat_schema.version):
            return cast(S, self._parse_obj(schema_class, obj.node[()]))
        return None

    def __setitem__(
        self, schema: Union[str, Type[S]], value: Union[Dict[str, Any], MetadataSchema]
    ) -> None:
        """Store metadata object as instance of given schema.

        Raises KeyError if passed schema is not installed in environment.

        Raises TypeError if passed schema is marked auxiliary.

        Raises ValueError if an object for the schema already exists.

        Raises ValidationError if passed object is not valid for the schema.
        """
        self._node._guard_acl(NodeAcl.read_only)
        schema_name, schema_ver = plugin_args(schema)

        # if self.get(schema_name, schema_ver):  # <- also subclass schemas
        # NOTE: for practical reasons let's be more lenient here and allow redundancy
        # hence only check if exact schema (modulo version) is already there
        if self._get_raw(schema_name):  # <- only same schema
            msg = f"Metadata object for schema {schema_name} already exists!"
            raise ValueError(msg)

        schema_class = self._require_schema(schema_name, schema_ver)
        checked_obj = self._parse_obj(schema_class, value)
        self._set_raw(schema_class.Plugin.ref(), checked_obj)

    def __delitem__(self, schema: Union[str, Type[MetadataSchema]]) -> None:
        """Delete metadata object explicitly stored for the passed schema.

        If a schema class is passed, its version is ignored,
        as each node may contain at most one explicit instance per schema.

        Raises KeyError if no metadata object for that schema exists.
        """
        self._node._guard_acl(NodeAcl.read_only)
        schema_name, _ = plugin_args(schema)

        if self._get_raw(schema_name) is None:
            raise KeyError(schema_name)  # no (explicit) metadata object

        self._del_raw(schema_name)


# ----


class TOCLinks:
    """Link management for synchronizing metadata objects and container TOC."""

    # NOTE: This is not exposed to the end-user

    @staticmethod
    def _link_path_for(schema_ref: PluginRef) -> str:
        return f"{M.METADOR_LINKS_PATH}/{_ep_name_for(schema_ref)}"

    def __init__(self, raw_cont: H5FileLike, toc_schemas: TOCSchemas):
        self._raw: H5FileLike = raw_cont
        """Raw underlying container (for quick access)."""

        self._toc_schemas = toc_schemas
        """Schemas used in container (to (un)register)."""

        self._toc_path: Dict[UUID, str] = {}
        """Maps metadata object UUIDs to paths of respective pseudo-symlink in TOC."""

        # load links into memory
        if M.METADOR_LINKS_PATH in self._raw:
            link_grp = self._raw.require_group(M.METADOR_LINKS_PATH)
            assert isinstance(link_grp, H5GroupLike)
            for schema_link_grp in link_grp.values():
                assert isinstance(schema_link_grp, H5GroupLike)
                for uuid, link_node in schema_link_grp.items():
                    assert isinstance(link_node, H5DatasetLike)
                    self._toc_path[UUID(uuid)] = link_node.name

    def fresh_uuid(self) -> UUID:
        """Return a UUID string not used for a metadata object in the container yet."""
        fresh = False
        ret: UUID
        # NOTE: here a very unlikely race condition is present if parallelized
        while not fresh:
            ret = uuid1()
            fresh = ret not in self._toc_path
        self._toc_path[ret] = None  # not assigned yet, but "reserved"
        # ----
        return ret

    def resolve(self, uuid: UUID) -> str:
        """Get the path a UUID in the TOC points to."""
        link_path = self._toc_path[uuid]
        link_node = cast(H5DatasetLike, self._raw[link_path])
        return link_node[()].decode("utf-8")

    def update(self, uuid: UUID, new_target: str):
        """Update target of an existing link to point to a new location."""
        link_path = self._toc_path[uuid]
        del self._raw[link_path]
        self._raw[link_path] = new_target

    def register(self, obj: StoredMetadata) -> None:
        """Create a link for a metadata object in container TOC.

        The link points to the metadata object.
        """
        self._toc_schemas._register(obj.schema)

        toc_path = f"{self._link_path_for(obj.schema)}/{obj.uuid}"
        self._toc_path[obj.uuid] = toc_path
        self._raw[toc_path] = str(obj.node.name)

    def unregister(self, uuid: UUID) -> None:
        """Unregister metadata object in TOC given its UUID.

        Will remove the object and clean up empty directories in the TOC.
        """
        # delete the link itself and free the UUID
        toc_path = self._toc_path[uuid]

        schema_group = self._raw[toc_path].parent
        assert isinstance(schema_group, H5GroupLike)
        link_group = schema_group.parent
        assert link_group.name == M.METADOR_LINKS_PATH

        del self._raw[toc_path]
        del self._toc_path[uuid]
        if len(schema_group):
            return  # schema still has instances

        s_name_vers: str = schema_group.name.split("/")[-1]
        # delete empty group for schema
        del self._raw[schema_group.name]
        # notify schema manager (cleans up schema + package info)
        self._toc_schemas._unregister(_schema_ref_for(s_name_vers))

        if len(link_group.keys()):
            return  # container still has metadata
        else:
            # remove the link dir itself (no known metadata in container left)
            del self._raw[link_group.name]

    # ----

    def find_broken(self, repair: bool = False) -> List[UUID]:
        """Return list of UUIDs in TOC not pointing to an existing metadata object.

        Will use loaded cache of UUIDs and check them, without scanning the container.

        If repair is set, will remove those broken links.
        """
        broken = []
        for uuid in self._toc_path.keys():
            target = self.resolve(uuid)
            if target not in self._raw:
                broken.append(uuid)
        if repair:
            for uuid in broken:
                self.unregister(uuid)
        return broken

    def find_missing(self, path: H5GroupLike) -> List[H5DatasetLike]:
        """Return list of metadata objects not listed in TOC."""
        missing = []

        def collect_missing(_, node):
            if not M.is_internal_path(node.name, M.METADOR_META_PREF):
                return  # not a metador metadata path
            if M.is_meta_base_path(node.name):
                # top dir, not a "link dataset",
                # e.g. /.../foo/metador_meta_ or /.../metador_meta_foo
                return

            # now we assume we have a path to a metadata link object in the group
            obj = StoredMetadata.from_node(node)
            known = obj.uuid in self._toc_path
            # check UUID collision: i.e., used in TOC, but points elsewhere
            # (requires fixing up the name of this object / new UUID)
            # implies that THIS object IS missing in the TOC
            collision = known and self.resolve(obj.uuid) != node.name
            if not known or collision:
                missing.append(node)

        # ensure its a group and collect
        self._raw.require_group(path.name).visititems(collect_missing)
        return missing

    def repair_missing(
        self, missing: List[H5DatasetLike], update: bool = False
    ) -> None:
        """Repair links (objects get new UUIDs, unless update is true)."""
        # NOTE: needed for correct copy and move of nodes with their metadata
        for node in missing:
            obj = StoredMetadata.from_node(node)
            if update and obj.uuid in self._toc_path:
                # update target of existing link (e.g. for move)
                self.update(obj.uuid, node.name)
            else:
                # assign new UUID (e.g. for copy)
                # (copied metadata node refers to some other uuid in the name)
                obj.uuid = self.fresh_uuid()
                new_path = obj.to_path()
                # rename the metadata node to point to the new UUID
                self._raw.move(node.name, new_path)
                obj.node = cast(H5DatasetLike, self._raw[new_path])
                # register the object with the new UUID in the TOC
                self.register(obj)


class TOCSchemas:
    """Schema management for schemas used in the container.

    Interface is made to mimic PGSchema wherever it makes sense.
    """

    @classmethod
    def _schema_path_for(cls, s_ref: PluginRef) -> str:
        return f"{M.METADOR_SCHEMAS_PATH}/{to_ep_name(s_ref.name, s_ref.version)}"

    @classmethod
    def _jsonschema_path_for(cls, s_ref: PluginRef) -> str:
        return f"{cls._schema_path_for(s_ref)}/jsonschema.json"

    @staticmethod
    def _load_json(node: H5DatasetLike):
        return json.loads(node[()].decode("utf-8"))

    def _update_parents_children(
        self, schema_ref: PluginRef, parents: Optional[List[PluginRef]]
    ):
        if parents is None:  # remove schema
            for parent in self._parents[schema_ref]:
                if parent in self._schemas:
                    self._children[parent].remove(schema_ref)
                elif all(
                    (child not in self._schemas for child in self._children[parent])
                ):
                    del self._parents[parent]
                    del self._children[parent]
        else:  # add schema
            for i, parent in enumerate(parents):
                if parent not in self._parents:
                    self._parents[parent] = parents[: i + 1]
                if parent not in self._children:
                    self._children[parent] = set()
                if parent != schema_ref:
                    self._children[parent].add(schema_ref)

    def _register(self, schema_ref: PluginRef):
        """Notify that a schema is used in the container (metadata object is created/updated).

        If the schema has not been used before in the container, will store metadata about it.
        """
        if schema_ref in self._schemas:
            return  # nothing to do

        # store json schema
        schema_cls = schemas.get(schema_ref.name, schema_ref.version)
        jsonschema_dat = schema_cls.schema_json().encode("utf-8")
        jsonschema_path = self._jsonschema_path_for(schema_ref)
        self._raw[jsonschema_path] = jsonschema_dat

        # store parent schema refs
        compat_path = f"{self._schema_path_for(schema_ref)}/compat"
        parents = schemas.parent_path(schema_ref.name, schema_ref.version)
        parents_dat: bytes = json.dumps(list(map(lambda x: x.dict(), parents))).encode(
            "utf-8"
        )

        self._raw[compat_path] = parents_dat
        self._schemas.add(schema_ref)
        self._update_parents_children(schema_ref, parents)

        # add providing package (if no stored package provides it)
        if not self._pkgs._providers.get(schema_ref, []):
            env_pkg_info: PluginPkgMeta = schemas.provider(schema_cls.Plugin.ref())
            pkg_name_ver = (str(env_pkg_info.name), env_pkg_info.version)
            self._pkgs._register(pkg_name_ver, env_pkg_info)
            self._used[pkg_name_ver] = set()

        # update used schemas tracker for all packages providing this schema
        for pkg in self._pkgs._providers[schema_ref]:
            self._used[pkg].add(schema_ref)

    def _unregister(self, schema_ref: PluginRef):
        """Notify that a schema is not used at any container node anymore.

        If after that no schema of a listed dep package is used,
        this dependency will be removed from the container.
        """
        del self._raw[self._schema_path_for(schema_ref)]
        self._schemas.remove(schema_ref)
        self._update_parents_children(schema_ref, None)

        providers = set(self._pkgs._providers[schema_ref])
        for pkg in providers:
            pkg_used = self._used[pkg]
            if schema_ref in pkg_used:
                # remove schema from list of used schemas of pkg
                pkg_used.remove(schema_ref)
            if not len(pkg_used):
                # package not used anymore in container -> clean up
                self._pkgs._unregister(pkg)

        # remove schemas group if it is empty (no schemas used in container)
        if not self._raw.require_group(M.METADOR_SCHEMAS_PATH).keys():
            del self._raw[M.METADOR_SCHEMAS_PATH]

    def __init__(self, raw_cont: H5FileLike, toc_packages: TOCPackages):
        self._raw: H5FileLike = raw_cont
        """Raw underlying container (for quick access)."""

        self._pkgs = toc_packages
        """TOC package metadata manager object."""

        self._schemas: Set[PluginRef] = set()
        """Stored JSON Schemas of actually used schemas."""

        self._parents: Dict[PluginRef, List[PluginRef]] = {}
        """Parents of a used json schema (i.e. other partially compatible schemas)."""

        self._children: Dict[PluginRef, Set[PluginRef]] = {}
        """Children of a used json schema (i.e. other fully compatible schemas)."""

        self._used: Dict[PythonDep, Set[PluginRef]] = {}
        """package name + version -> name of schemas used in container"""

        for pkg in self._pkgs.keys():
            self._used[pkg] = set()

        if M.METADOR_SCHEMAS_PATH in self._raw:
            schema_grp = self._raw.require_group(M.METADOR_SCHEMAS_PATH)
            for name, node in schema_grp.items():
                s_ref: PluginRef = _schema_ref_for(name)
                assert isinstance(node, H5GroupLike)
                compat = node["compat"]
                assert isinstance(compat, H5DatasetLike)

                reflist = json.loads(compat[()].decode("utf-8"))
                parents = list(map(PluginRef.parse_obj, reflist))

                self._schemas.add(s_ref)
                self._update_parents_children(s_ref, parents)
                for pkg in self._pkgs._providers[s_ref]:
                    self._used[pkg].add(s_ref)

    @property
    def packages(self) -> TOCPackages:
        """Like PluginGroup.packages, but with respect to schemas used in container."""
        return self._pkgs

    def provider(self, schema_ref: PluginRef) -> PluginPkgMeta:
        """Like PluginGroup.provider, but with respect to container deps."""
        pkg_name_ver = next(iter(self._pkgs._providers.get(schema_ref, [])), None)
        if pkg_name_ver is None:
            msg = f"Did not find metadata of a package providing schema: '{schema_ref}'"
            raise KeyError(msg)
        return self._pkgs[pkg_name_ver]

    def parent_path(
        self, schema, version: Optional[SemVerTuple] = None
    ) -> List[PluginRef]:
        """Like PGSchema.parent_path, but with respect to container deps."""
        name, vers = plugin_args(schema, version, require_version=True)
        s_ref = schemas.PluginRef(name=name, version=vers)
        return self._parents[s_ref]

    def versions(
        self, p_name: str, version: Optional[SemVerTuple] = None
    ) -> List[PluginRef]:
        """Like PGSchema.versions, but with respect to container deps."""
        # NOTE: using _children instead of _schemas because some are only listed
        # due to their appearance in the parent_path of some actually used schema
        # but we need them here for "parent compatibility" to work right.
        refs = list(filter(lambda s: s.name == p_name, self._children))

        if version is None:
            return refs
        # filter plugins for compatible version
        requested = schemas.PluginRef(name=p_name, version=version)
        # NOTE: here "supports" arguments are reversed (compared to "plugin versions")!
        # because its about instances (that must be "below" the requested schema version)
        return [ref for ref in refs if requested.supports(ref)]

    def children(self, schema, version: Optional[SemVerTuple] = None) -> Set[PluginRef]:
        """Like PGSchema.children, but with respect to container deps."""
        name, vers = plugin_args(schema, version)
        if vers is not None:
            s_refs = [schemas.PluginRef(name=name, version=vers)]
        else:
            # if no version is given, collect all possibilities
            s_refs = [ref for ref in self._children.keys() if ref.name == name]
        # return all that can be actually retrieved
        return set().union(
            *filter(lambda x: x is not None, map(self._children.get, s_refs))
        )

    # ----

    def __len__(self):
        return len(self._schemas)

    def __iter__(self):
        return iter(self.keys())

    def __contains__(self, schema_ref: PluginRef):
        return schema_ref in self._schemas

    def __getitem__(self, schema_ref: PluginRef):
        node_path = self._jsonschema_path_for(schema_ref)
        assert node_path in self._raw
        return self._load_json(cast(H5DatasetLike, self._raw[node_path]))

    def get(self, schema_ref: PluginRef):
        try:
            self[schema_ref]
        except KeyError:
            return None

    def keys(self):
        return set(self._schemas)

    def values(self):
        return [self[k] for k in self.keys()]

    def items(self):
        return [(k, self[k]) for k in self.keys()]


PythonDep: TypeAlias = Tuple[str, SemVerTuple]


class TOCPackages:
    """Package metadata management for schemas used in the container.

    The container will always store for each schema used in the
    information about one package providing that schema.

    If there are multiple providers of the same schema,
    the first/existing one is preferred.
    """

    @staticmethod
    def _pkginfo_path_for(pkg_name: str, pkg_version: SemVerTuple) -> str:
        return f"{M.METADOR_PACKAGES_PATH}/{to_ep_name(pkg_name, pkg_version)}"

    def _add_providers(self, pkg: PythonDep, pkginfo: PluginPkgMeta):
        # fill schema -> package lookup table for provided package
        for schema_ref in pkginfo.plugins[schemas.name]:
            if schema_ref not in self._providers:
                self._providers[schema_ref] = set()
            self._providers[schema_ref].add(pkg)

    def _register(self, pkg: PythonDep, info: PluginPkgMeta):
        pkg_path = self._pkginfo_path_for(*pkg)
        self._raw[pkg_path] = bytes(info)
        self._pkginfos[pkg] = info
        self._add_providers(pkg, info)

    def _unregister(self, pkg: PythonDep):
        pkg_path = self._pkginfo_path_for(*pkg)
        del self._raw[pkg_path]
        info = self._pkginfos.pop(pkg)
        # unregister providers
        for schema_ref in info.plugins[schemas.name]:
            providers = self._providers[schema_ref]
            providers.remove(pkg)
            if not providers:  # schema not provided by any package
                del self._providers[schema_ref]

        # remove schemas group if it is empty (no schemas used in container)
        if not self._raw.require_group(M.METADOR_PACKAGES_PATH).keys():
            del self._raw[M.METADOR_PACKAGES_PATH]

    def __init__(self, raw_container: H5FileLike):
        self._raw: H5FileLike = raw_container
        """Raw underlying container (for quick access)."""

        self._pkginfos: Dict[PythonDep, PluginPkgMeta] = {}
        """Package name + version -> package info"""

        self._providers: Dict[PluginRef, Set[PythonDep]] = {}
        """schema reference -> package name + version"""

        # parse package infos if they exist
        if M.METADOR_PACKAGES_PATH in self._raw:
            deps_grp = self._raw.require_group(M.METADOR_PACKAGES_PATH)
            for name, node in deps_grp.items():
                pkg: PythonDep = from_ep_name(EPName(name))
                info = PluginPkgMeta.parse_raw(cast(H5DatasetLike, node)[()])
                self._pkginfos[pkg] = info
                self._add_providers(pkg, info)

    # ----

    def __len__(self):
        return len(self._pkginfos)

    def __iter__(self):
        return iter(self._pkginfos)

    def __contains__(self, pkg: PythonDep):
        return pkg in self._pkginfos

    def __getitem__(self, pkg: PythonDep):
        return self._pkginfos[pkg]

    def keys(self):
        return self._pkginfos.keys()

    def values(self):
        return self._pkginfos.values()

    def items(self):
        return self._pkginfos.items()


class MetadorContainerTOC:
    """Interface to the Metador metadata index (table of contents) of a container."""

    def __init__(self, container: MetadorContainer):
        self._container = container
        self._raw = self._container.__wrapped__

        ver = self.spec_version if M.METADOR_VERSION_PATH in self._raw else None
        if ver:
            if ver >= [2]:
                msg = f"Unsupported Metador container version: {ver}"
                raise ValueError(msg)
        else:
            if self._container.acl[NodeAcl.read_only]:
                msg = "Container is read-only and does not look like a Metador container! "
                msg += "Please open in writable mode to initialize Metador structures!"
                raise ValueError(msg)

            # writable + no version = fresh (for metador), initialize it
            self._raw[M.METADOR_VERSION_PATH] = M.METADOR_SPEC_VERSION
            self._raw[M.METADOR_UUID_PATH] = str(uuid1())

        # if we're here, we have a prepared container TOC structure

        # proceed to initialize TOC
        self._driver_type: MetadorDriverEnum = get_driver_type(self._raw)

        self._packages = TOCPackages(self._raw)
        self._schemas = TOCSchemas(self._raw, self._packages)
        self._links = TOCLinks(self._raw, self._schemas)

    # ----

    @property
    def driver_type(self) -> MetadorDriverEnum:
        """Return the type of the container driver."""
        return self._driver_type

    @property
    def driver(self) -> Type[MetadorDriver]:
        """Return the container driver class used by the container."""
        return METADOR_DRIVERS[self.driver_type]

    @property
    def source(self) -> Any:
        """Return data underlying thes container (file, set of files, etc. used with the driver)."""
        return get_source(self._raw, self.driver_type)

    # ----

    @property
    def container_uuid(self) -> UUID:
        """Return UUID of the container."""
        uuid = self._raw[M.METADOR_UUID_PATH]
        uuid_ds = cast(H5DatasetLike, uuid)
        return UUID(uuid_ds[()].decode("utf-8"))

    @property
    def spec_version(self) -> List[int]:
        """Return Metador container specification version of the container."""
        ver = cast(H5DatasetLike, self._raw[M.METADOR_VERSION_PATH])
        return list(map(int, ver[()].decode("utf-8").split(".")))

    @property
    def schemas(self):
        """Information about all schemas used for metadata objects in this container."""
        return self._schemas

    def query(
        self,
        schema: Union[str, Type[S]],
        version: Optional[SemVerTuple] = None,
        *,
        node: Optional[MetadorNode] = None,
    ) -> Iterator[MetadorNode]:
        """Return nodes that contain a metadata object compatible with the given schema."""
        schema_name, schema_ver = plugin_args(schema, version)
        if not schema_name:  # could be e.g. empty string
            msg = "A schema name, plugin reference or class must be provided!"
            raise ValueError(msg)

        start_node: MetadorNode = node or self._container["/"]

        # check start node metadata explicitly
        if (schema_name, schema_ver) in start_node.meta:
            yield start_node

        if not isinstance(start_node, H5GroupLike):
            return  # the node is not group-like, cannot be traversed down

        # collect nodes below start node recursively
        # NOTE: yielding from the collect_nodes does not work :'(
        # so we have to actually materialize the list >.<
        # but we expose only the generator interface anyway (better design)
        # (maybe consider replacing visititems with a custom traversal here)
        ret: List[MetadorNode] = []

        def collect_nodes(_, node: MetadorNode):
            if (schema_name, schema_ver) in node.meta:
                ret.append(node)

        start_node.visititems(collect_nodes)
        yield from iter(ret)



import pytest

from metador_core.container import MetadorContainer
from metador_core.container.drivers import MetadorDriverEnum


@pytest.fixture(scope="function", params=list(iter(MetadorDriverEnum)))
def mc_driver(request):
    """Provide different metador container driver enum values."""
    return request.param


@pytest.fixture(scope="function")
def tmp_mc_path(tmp_ds_path):
    """Return a temporary path suitable for MetadorContainer creation."""
    tmp_ds_path.mkdir()
    return tmp_ds_path / "container"


@pytest.fixture
def fresh_raw(tmp_mc_path, mc_driver):
    # will not autoclose
    return mc_driver.value(tmp_mc_path, "w")


@pytest.fixture
def fresh_raw_managed(fresh_raw):
    # will auto-close
    with fresh_raw:
        yield fresh_raw


@pytest.fixture(scope="function")
def fresh_mc(fresh_raw):
    """Return fresh writable MetadorContainer instance.

    The dataset will be closed and destroyed afterwards.
    """
    with MetadorContainer(fresh_raw) as m:
        yield m



"""Example GenericPacker tests."""

import h5py
import pytest

from metador_core.packer.utils import DirValidationErrors
from metador_core.plugins import packers

pytest.skip(reason="FIXME when API complete", allow_module_level=True)


def test_example_packer(tmp_ds_path, testinputs):
    ds1 = testinputs("dirdiff1")
    ds2 = testinputs("dirdiff1")

    generic = packers["generic"]

    # prepare directory and check that it is packer-compatible (just to make sure)
    assert not generic.check_dir(ds1)

    # rename a file to make it fail
    (ds1 / "_meta.yaml").rename(ds1 / "renamed")
    errs = generic.check_dir(ds1)
    assert errs
    print(errs)

    with pytest.raises(DirValidationErrors):
        packers.pack("generic", ds1, tmp_ds_path, h5py.File)

    # fix error with directory
    (ds1 / "renamed").rename(ds1 / "_meta.yaml")

    # do some "changes"
    with open(ds2 / "example_meta.yaml", "w") as f:
        f.write("author: changed")

    # now update
    packers.update("generic", ds2, tmp_ds_path, h5py.File)



"""Interface of Metador widget plugins."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    Type,
)

from overrides import overrides
from panel.viewable import Viewable
from pydantic import Field
from typing_extensions import Annotated, TypeAlias

from ..container import MetadorDataset, MetadorNode
from ..plugin import interface as pg
from ..plugins import schemas
from ..schema import MetadataSchema
from ..schema.plugins import PluginRef
from ..schema.types import SemVerTuple
from .server import WidgetServer


class Widget(ABC):
    """Base class for metador widgets."""

    _args: Dict[str, Any]
    """Additional passed arguments (e.g. from the dashboard)."""

    _node: MetadorNode
    """Container node passed to the widget."""

    _meta: MetadataSchema
    """Metadata object to be used as the starting point (e.g. widget config)."""

    _server: WidgetServer
    """Widget-backing server object to use (e.g. to access container files from frontend)."""

    Plugin: ClassVar[WidgetPlugin]

    def __init__(
        self,
        node: MetadorNode,
        schema_name: str = "",
        schema_version: Optional[SemVerTuple] = None,
        *,
        server: Optional[WidgetServer] = None,
        container_id: Optional[str] = None,
        metadata: Optional[MetadataSchema] = None,
        max_width: Optional[int] = None,
        max_height: Optional[int] = None,
        keep_ratio: bool = False,
    ):
        """Instantiate a widget for a node.

        If no schema name is provided, the widget will try to pick the first metadata object
        from the node that is an instance of a supported schema, in the listed order.

        If no server is provided, a stand-alone server is started (e.g. for use in a notebook).

        If a metadata object is passed explicitly, it will be used instead of trying to
        retrieve one from the node.
        """
        # NOTE: we restrict the node so that widgets don't try to escape their scope
        self._node = node.restrict(read_only=True, local_only=True)
        # NOTE: if no container_id is passed, we assume its jupyter mode (based on the metador UUIDs)
        self._container_id = container_id or str(node.metador.container_uuid)

        # if no server passed, we're in Jupyter mode - use standalone
        srv: WidgetServer
        if server is not None:
            srv = server
        else:
            from .jupyter.standalone import running, widget_server

            if not running():
                raise ValueError(
                    "No widget server passed and standalone server not running!"
                )
            srv = widget_server()
        self._server = srv

        # setup correct metadata
        if metadata is not None:
            if not self.supports_meta(metadata):
                msg = "Passed metadata is not instance of a supported schema!"
                raise ValueError(msg)
            self._meta = metadata
        else:
            if not schema_name:
                for schemaref in self.Plugin.supports:
                    if node.meta.get(schemaref.name):
                        schema_name = schemaref.name
                        break
            if not schema_name:
                raise ValueError("The node does not contain any suitable metadata!")

            if metadata := node.meta.get(schema_name, schema_version):
                self._meta = metadata
            else:
                raise ValueError("The node does not contain '{schema_name}' metadata!")

        # maximum width and height that can be used (if None, unlimited)
        self._w: Optional[int] = max_width
        self._h: Optional[int] = max_height

        # recalibrate maximum width and height of widgets to preserve ration, if possible + desired
        can_scale = self._meta.width is not None and self._meta.height is not None
        if keep_ratio and can_scale:
            scale_factor = min(
                self._h / self._meta.height.value, self._w / self._meta.width.value
            )
            self._w = int(self._meta.width.value * scale_factor)
            self._h = int(self._meta.height.value * scale_factor)

        # widget-specific setup hook
        self.setup()

    def file_data(self, node: Optional[MetadorDataset] = None) -> bytes:
        """Return data at passed dataset node as bytes.

        If no node passed, will use the widget root node (if it is a dataset).
        """
        node = node or self._node
        if not isinstance(node, MetadorDataset):
            raise ValueError(
                f"Passed node {node.name} does not look like a dataset node!"
            )
        return node[()].tolist()

    def file_url(self, node: Optional[MetadorNode] = None) -> str:
        """Return URL resolving to the data at given node.

        If no node passed, will use the widget root node (if it is a dataset).
        """
        node = node or self._node
        if not isinstance(node, MetadorDataset):
            raise ValueError(
                f"Passed node {node.name} does not look like a dataset node!"
            )
        return self._server.file_url_for(self._container_id, node)

    @classmethod
    def supports(cls, *schemas: PluginRef) -> bool:
        """Return whether any (exact) schema is supported (version-compatible) by widget."""
        return any(
            any(map(lambda sref: sref.supports(schema), cls.Plugin.supports))
            for schema in schemas
        )

    @classmethod
    def supports_meta(cls, obj: MetadataSchema) -> bool:
        """Return whether widget supports the specific metadata object.

        The passed object is assumed to be of one of the supported schema types.

        Default implementation will just check that the object is of a supported schema.

        Override to constrain further (e.g. check field values).

        This method affects the dashboard widget selection process and is used
        to check a metadata object if directly passed to `__init__`.
        """
        return cls.supports(type(obj).Plugin.ref())

    def setup(self):  # noqa: B027  # implementing it is not mandatory
        """Check that passed node is valid and do preparations.

        If multiple supported schemas are listed, case splitting based on the
        schema type should be done here to minimize logic in the rendering.

        Everything that instances can reuse, especially if it is computationally
        expensive, should also be done here.

        In case the widget is not able to work with the given node and metadata,
        it will raise a `ValueError`.
        """

    @abstractmethod
    def show(self) -> Viewable:
        """Return a fresh Panel widget representing the node data and/or metadata.

        If width and height were provided during initialization, the widget is supposed
        to fit within these dimensions, not exceed them and if possible, usefully
        fill up the space.

        This method assumes that the widget is fully initialized and setup is completed.
        """
        raise NotImplementedError


WIDGET_GROUP_NAME = "widget"

if TYPE_CHECKING:
    SchemaPluginRef: TypeAlias = PluginRef
else:
    SchemaPluginRef = schemas.PluginRef


class WidgetPlugin(pg.PluginBase):
    supports: Annotated[List[SchemaPluginRef], Field(min_items=1)]  # type: ignore
    """Return list of schemas supported by this widget."""

    primary: bool = True
    """Return whether the widget is a primary choice.

    If False, will not be used automatically by dashboard.
    """


class PGWidget(pg.PluginGroup[Widget]):
    """Widget plugin group interface."""

    class Plugin:
        name = WIDGET_GROUP_NAME
        version = (0, 1, 0)

        requires = [PluginRef(group="plugingroup", name="schema", version=(0, 1, 0))]

        plugin_class = Widget
        plugin_info_class = WidgetPlugin

    @overrides
    def check_plugin(self, ep_name: str, plugin: Type[Widget]):
        pg.util.check_implements_method(ep_name, plugin, Widget.show)

    def plugin_deps(self, plugin) -> Set[PluginRef]:
        return set(plugin.Plugin.supports)

    def widgets_for(self, schema: PluginRef) -> Iterator[PluginRef]:
        """Return widgets that support (a parent of) the given schema."""
        ws = set()
        p_path = schemas.parent_path(schema.name, schema.version)
        for s_ref in reversed(p_path):  # in decreasing specifity
            for w_cls in self.values():
                if w_cls.supports(s_ref) and s_ref not in ws:
                    w_ref = w_cls.Plugin.ref()
                    ws.add(w_ref)
                    yield w_ref

    # def supported_schemas(self) -> Set[PluginRef]:
    #     """Return union of all schemas supported by all installed widgets.

    #     This includes registered child schemas (subclasses of supported schemas).
    #     """
    #     supported = set()
    #     for w in self.values():
    #         for sref in w.Plugin.supports:
    #             supported.add(sref)
    #             for cs_ref in schemas.children(sref.name, sref.version):
    #                 supported.add(cs_ref)
    #     return supported



"""Test plugingroup import mechanism."""
import pytest


def test_plugin_module_exports():
    # check what the module lists as available
    from metador_core import plugins

    assert dir(plugins) == plugins.__all__
    assert "plugingroups" in plugins.__all__
    assert "schemas" in plugins.__all__


def test_plugin_import():
    with pytest.raises(ImportError):
        # not a valid plugingroup
        from metador_core.plugins import foos  # noqa: F401
    with pytest.raises(ImportError):
        # not a plural (must have a suffix 's')
        from metador_core.plugins import schema  # noqa: F401

    # should work
    from metador_core.plugins import plugingroups, schemas  # noqa: F401



"""RO Crate compatible Metadata.

Here we do impose certain constraints (make fields mandatory).

See https://www.researchobject.org/ro-crate/1.1/
"""
from __future__ import annotations

from typing import Set

from pydantic import parse_obj_as, root_validator, validator

from ..decorators import make_mandatory
from ..ld import LDIdRef, ld_decorator
from ..types import MimeTypeStr, NonEmptyStr
from . import schemaorg

CTX_URL_ROCRATE = "https://w3id.org/ro/crate/1.1/context"


rocrate = ld_decorator(context=CTX_URL_ROCRATE)


@make_mandatory("contentSize", "sha256")
@rocrate(type="File")
class FileMeta(schemaorg.MediaObject):
    class Plugin:
        name = "core.file"
        version = (0, 1, 0)

    # NOTE: We do not use `name` here because `name` is used semantically
    # like a title in schema.org, which could also make sense for a file to have.
    filename: NonEmptyStr
    """Original name of the file in source directory."""

    encodingFormat: MimeTypeStr
    """MIME type of the file."""


@rocrate(type="Dataset")
class DirMeta(schemaorg.Dataset):
    class Plugin:
        name = "core.dir"
        version = (0, 1, 0)

    hasPart: Set[LDIdRef] = set()
    """References to (a subset of) contained files and subdirectories."""


@rocrate(type="Organization")
class Organization(schemaorg.Organization):
    @validator("id_")
    def check_id(cls, v):
        if not v:
            return None
        if not v.startswith("https://ror.org/"):
            raise ValueError("Person @id must be a valid ROR URL!")
        return parse_obj_as(schemaorg.URL, v)


@rocrate(type="Person")
class Person(schemaorg.Person):
    @validator("id_")
    def check_id(cls, v):
        if not v:
            return None
        if not v.startswith("https://orcid.org/"):
            raise ValueError("Person @id must be a valid ORCID URL!")
        return parse_obj_as(schemaorg.URL, v)

    @root_validator(pre=True)
    def check_name(cls, values):
        has_given_or_additional = values.get("givenName") or values.get(
            "additionalName"
        )
        if has_given_or_additional and not values.get("familyName"):
            msg = "givenName and additionalName require also familyName to be provided!"
            raise ValueError(msg)
        return values

    @root_validator
    def ensure_name(cls, values):
        missing_name = values.get("name") is None
        if missing_name:
            parts = []
            for k in ["givenName", "additionalName", "familyName"]:
                if v := values.get(k):
                    parts.append(v)
            fullname = " ".join(parts)
            if not fullname:
                msg = "A Person must have name or familyName set!"
                raise ValueError(msg)
            values["name"] = fullname
        return values


# required self-description of RO-Crate file
rocrate_self_meta = {
    "@type": "CreativeWork",
    "@id": "ro-crate-metadata.json",
    "conformsTo": {"@id": "https://w3id.org/ro/crate/1.1"},
    "about": {"@id": "./"},
}



import numpy as np
import pytest

from metador_core.container.drivers import (
    MetadorDriverEnum,
    get_driver_type,
    get_source,
    to_h5filelike,
)
from metador_core.container.protocols import (
    H5DatasetLike,
    H5FileLike,
    H5GroupLike,
    H5NodeLike,
)


def check_protocols_for(f):
    # create group and dataset node
    f["group/dataset"] = 123

    # check that behavior is correct:

    # file-like:
    assert isinstance(f, H5NodeLike)
    assert isinstance(f, H5FileLike)
    assert isinstance(f, H5GroupLike)
    assert not isinstance(f, H5DatasetLike)

    # group-like:
    g = f["group"]
    assert isinstance(f, H5NodeLike)
    assert not isinstance(g, H5FileLike)
    assert isinstance(g, H5GroupLike)
    assert not isinstance(g, H5DatasetLike)

    # dataset-like:
    d = g["dataset"]
    assert isinstance(f, H5NodeLike)
    assert not isinstance(d, H5FileLike)
    assert not isinstance(d, H5GroupLike)
    assert isinstance(d, H5DatasetLike)


def test_instance_checks(fresh_mc):
    # Check that all driver classes work correctly with the protocol types.
    #
    # Expected relationships (< = base_class_of):
    #  H5DatasetLike > H5NodeLike < H5GroupLike < H5FileLike
    check_protocols_for(fresh_mc.__wrapped__)


def test_instance_checks_wrapped(fresh_mc):
    # same, but wrapped with MetadorContainer
    check_protocols_for(fresh_mc)


# ----


def test_driver_type_detection(tmp_mc_path, mc_driver):
    with mc_driver.value(tmp_mc_path, "w") as f:
        # check to get expected driver type
        assert get_driver_type(f) == mc_driver

    class sub_d_cls(mc_driver.value):
        ...

    with sub_d_cls(tmp_mc_path, "r") as f:
        # a subclass should be still same "driver type"
        assert get_driver_type(f) == mc_driver

    with pytest.raises(ValueError):
        # invalid argument type
        get_driver_type(123)


def test_driver_source_detection(tmp_mc_path, mc_driver):
    with mc_driver.value(tmp_mc_path, "w") as f:
        # extract data source object
        src = get_source(f, driver=mc_driver)

    with mc_driver.value(src, "r") as f:
        # can open it again using returned data source object
        pass

    with pytest.raises(ValueError):
        # invalid argument type
        get_source(123)


def test_to_h5filelike(tmp_ds_path):
    def_drv_cls = MetadorDriverEnum.HDF5.value

    tmp_ds_path.mkdir()

    with pytest.raises(ValueError):
        # unknown driver class
        to_h5filelike(tmp_ds_path / "some_file", driver=object)

    with def_drv_cls(tmp_ds_path / "hdf5_file", "w") as f:
        # returns back same object
        assert to_h5filelike(f) is f

    # opens object with default driver
    assert isinstance(to_h5filelike(tmp_ds_path / "hdf5_file"), def_drv_cls)


# ----
# test driver implementation, i.e. the methods in the protocols
# actually work as expected and needed for the Metador interface.
# rigorous testing should be done for each driver separately.


def test_h5filelike(tmp_mc_path, mc_driver):
    # 'with' notation should work (__enter__, __exit__)
    with mc_driver.value(tmp_mc_path, "w") as f:
        # new file is empty and writable
        assert len(f) == 0
        assert f.mode == "r+"
        # check that we can write a value as expected
        f["a"] = b"b"

    with mc_driver.value(tmp_mc_path, "r") as f:
        # existing file is readable and contains what it should
        assert len(f) == 1
        assert f.mode == "r"
        assert list(f.keys()) == ["a"]
        assert f["a"][()] == b"b"

    f = mc_driver.value(tmp_mc_path, "r")
    f.close()
    with pytest.raises(ValueError):
        f["a"]  # access to a closed file should not work


def test_h5datasetlike(fresh_raw_managed):
    # test simple datatypes that are important
    c = fresh_raw_managed

    i = c.create_dataset("int", data=123)
    assert i.ndim == 0
    assert i[()] == 123  # __getattr__

    s = c.create_dataset("str", data="hello")
    assert s.ndim == 0
    assert s[()] == b"hello"

    b = c.create_dataset("bytes", data=b"world")
    assert b.ndim == 0
    assert b[()] == b"world"

    v = c.create_dataset("void", data=np.void(b"wrapped"))
    assert v.ndim == 0
    assert v[()].tolist() == b"wrapped"

    lst = c.create_dataset("mat", data=[[1, 2], [3, 4]])
    assert lst.ndim == 2
    assert lst[1, 1] == 4
    lst[1, 1] = 5  # __setattr__, nontrivial indexing
    assert lst[1, 1] == 5


def test_h5nodelike(fresh_raw_managed):
    f = fresh_raw_managed
    f["a/b"] = 123
    g = f["a"]
    d = g["b"]

    assert f.name == "/"
    assert g.name == "/a"
    assert d.name == "/a/b"

    assert d.parent == g
    assert g.parent == f

    assert g.file == f
    assert d.file == f

    # TODO: should it be checked more than that attrs exist?
    # we don't use them but packers might.
    # depends on whether packers should be driver agnostic
    assert f.attrs is not None
    assert d.attrs is not None


def test_h5grouplike(fresh_raw_managed):
    f = fresh_raw_managed

    assert "a" not in f
    assert "a/b" not in f
    assert f.get("a/b/c") is None

    f["a/b"] = 123  # __setattr__
    g = f["a"]  # __getattr__
    d = g.get("b")
    assert d[()] == 123
    g2 = g.create_group("d")
    g2.create_dataset("/a/c", data=456)
    assert len(g2) == 0

    tmp_d = g2.require_dataset("e", shape=(10, 10), dtype="float64")
    assert isinstance(tmp_d, H5DatasetLike)
    tmp_g = g2.require_group("f")
    assert isinstance(tmp_g, H5GroupLike)
    with pytest.raises(TypeError):
        g2.require_group("e")
    with pytest.raises(TypeError):
        g2.require_dataset("f", shape=(10, 10), dtype="float64")

    assert len(g2) == 2
    del g2["e"]  # __delitem__
    del g["/a/d/f"]
    assert len(g2) == 0

    assert list(f.values()) == [g]
    assert dict(f.items()) == {"a": g}
    assert list(f.keys()) == ["a"]
    assert list(iter(f)) == ["a"]
    assert list(g.keys()) == ["b", "c", "d"]

    assert "/a" in f
    assert "a/b" in f
    assert "b" not in f
    assert "b" in g
    assert "/a" in g
    assert "/b" not in g
    assert "/a/b" in g

    assert f["a/b"] == f["/a/b"]
    assert f["a/b"] == g["b"]
    assert f["/"] == g["/"]

    g.move("/a", "/g")
    f.move("g/b", "b")
    f["g"].move("c", "/c")
    f.move("/g", "a")
    assert list(f.keys()) == ["a", "b", "c"]
    assert list(f["a"].keys()) == ["d"]

    f.copy("b", f["a/d"])
    f.copy("c", "/a/d/c")
    f["a"].copy(f["a/d"], f)
    assert list(f.keys()) == ["a", "b", "c", "d"]
    assert list(f["d"].keys()) == ["b", "c"]
    assert f["d/b"][()] == 123

    collected = []

    def collect(name):
        collected.append(name)

    f["a"].visit(collect)
    assert collected == ["d", "d/b", "d/c"]

    collected = []

    def collect(name, node):
        assert node.name == f"/a/{name}"
        collected.append(name)

    f["a"].visititems(collect)
    assert collected == ["d", "d/b", "d/c"]



"""Constants and helper functions.

Provides syntactic path transformations to implement the Metador container layout.
"""

from typing_extensions import Final

METADOR_SPEC_VERSION: Final[str] = "1.0"
"""Version of container spec created by this package."""
# NOTE: don't forget to change it when something about the container structure changes!

METADOR_PREF: Final[str] = "metador_"
"""Reserved prefix for group and dataset names."""

METADOR_META_PREF: Final[str] = METADOR_PREF + "meta_"
"""Sub-prefix for group that stores group or dataset metadata."""

METADOR_TOC_PATH: Final[str] = f"/{METADOR_PREF}container"
"""Path of group with the Metador metadata index structure of the container."""

METADOR_VERSION_PATH: Final[str] = f"{METADOR_TOC_PATH}/version"
"""Path of dataset with the Metador container spec version of the container."""

METADOR_UUID_PATH: Final[str] = f"{METADOR_TOC_PATH}/uuid"
"""Path of dataset with the Metador container version of the container."""

METADOR_PACKAGES_PATH: Final[str] = f"{METADOR_TOC_PATH}/packages"
"""Path of group with package info of packages providing used schemas in the container."""

METADOR_SCHEMAS_PATH: Final[str] = f"{METADOR_TOC_PATH}/schemas"
"""Path of group with info about used schemas in the container."""

METADOR_LINKS_PATH: Final[str] = f"{METADOR_TOC_PATH}/links"
"""Path of group with links to schema instances in the container."""


def is_internal_path(path: str, pref: str = METADOR_PREF) -> bool:
    """Return whether the path of this node is Metador-internal (metador_*).

    Optional argument can set a different prefix that path segments are checked for.
    """
    # first case is for relative paths, second for later path segments and absolute paths
    return path.startswith(pref) or path.find(f"/{pref}") >= 0


# filtering and transforming between node and metadata base path


def is_meta_base_path(path: str) -> bool:
    """Return whether the path is a metadata base dir (but not an inner path!)."""
    return path.split("/")[-1].startswith(METADOR_META_PREF)


def to_meta_base_path(node_path: str, is_dataset: bool) -> str:
    """Return path to base group containing metadata for given node."""
    segs = node_path.split("/")
    if is_dataset:
        segs[-1] = METADOR_META_PREF + segs[-1]
    elif segs == ["", ""]:  # name was "/"
        segs[-1] = METADOR_META_PREF
    else:
        segs.append(METADOR_META_PREF)
    return "/".join(segs)


def to_data_node_path(meta_dir_path: str) -> str:
    """Given a metadata group path, infer the correct node path.

    Path can be relative or absolute.
    Will not check validity of the passed path, assumes it is fitting the scheme!
    """
    segs = meta_dir_path.split("/")
    pl = len(METADOR_META_PREF)
    segs[-1] = segs[-1][pl:]
    if segs[-1] == "" and (len(segs) > 2 or segs[0] != ""):
        segs.pop()
    return "/".join(segs)



"""Interface for plugin groups."""
from __future__ import annotations

from abc import ABCMeta
from typing import (
    Any,
    ClassVar,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from importlib_metadata import EntryPoint
from typing_extensions import TypeAlias

from ..schema.plugins import PluginBase, PluginPkgMeta
from ..schema.plugins import PluginRef as AnyPluginRef
from ..util import eprint
from . import util
from .entrypoints import get_group, pkg_meta
from .metaclass import UndefVersion
from .types import (
    EP_NAME_REGEX,
    EPName,
    PluginLike,
    SemVerTuple,
    ep_name_has_namespace,
    from_ep_name,
    is_pluginlike,
    plugin_args,
    to_semver_str,
)

PG_GROUP_NAME = "plugingroup"


class PGPlugin(PluginBase):
    plugin_info_class: Optional[Type[PluginBase]] = None
    plugin_class: Optional[Any] = object


# TODO: plugin group inheritance is not checked yet because it adds complications
class PluginGroupMeta(ABCMeta):
    """Metaclass to initialize some things on creation."""

    def __init__(self, name, bases, dct):
        assert is_pluginlike(self, check_group=False)

        # attach generated subclass that auto-fills the group for plugin infos
        self.PluginRef: Type[AnyPluginRef] = AnyPluginRef._subclass_for(
            self.Plugin.name
        )

        if pgi_cls := self.Plugin.__dict__.get("plugin_info_class"):
            # attach group name to provided plugin info class
            pgi_cls.group = self.Plugin.name
        else:
            # derive generic plugin info class with the group defined
            class PGInfo(PluginBase):
                group = self.Plugin.name

            self.Plugin.plugin_info_class = PGInfo

        # sanity checks... this magic should not mess with the PluginBase
        assert self.Plugin.plugin_info_class is not PluginBase
        assert PluginBase.group == ""


T = TypeVar("T", bound=PluginLike)


class PluginGroup(Generic[T], metaclass=PluginGroupMeta):
    """All pluggable entities in metador are subclasses of this class.

    The type parameter is the (parent) class of all loaded plugins.

    They must implement the check method and be listed as plugin group.
    The name of their entrypoint defines the name of the plugin group.
    """

    PluginRef: TypeAlias = AnyPluginRef
    """Plugin reference class for this plugin group."""

    _PKG_META: ClassVar[Dict[str, PluginPkgMeta]] = pkg_meta
    """Package name -> package metadata."""

    class Plugin:
        """This is the plugin group plugin group, the first loaded group."""

        name = PG_GROUP_NAME
        version = (0, 1, 0)
        plugin_info_class = PGPlugin
        plugin_class: Type
        # plugin_class = PluginGroup  # can't set that -> check manually

    _ENTRY_POINTS: Dict[EPName, EntryPoint]
    """Dict of entry points of versioned plugins (not loaded)."""

    _VERSIONS: Dict[str, List[AnyPluginRef]]
    """Mapping from plugin name to pluginrefs of available versions."""

    _LOADED_PLUGINS: Dict[AnyPluginRef, Type[T]]
    """Dict from entry points to loaded plugins of that pluggable type."""

    def _add_ep(self, epname_str: str, ep_obj: EntryPoint):
        """Add an entrypoint loaded from importlib_metadata."""
        try:
            ep_name = EPName(epname_str)
        except TypeError:
            msg = f"{epname_str}: Invalid entrypoint name, must match {EP_NAME_REGEX}"
            raise ValueError(msg)
        if type(self) is not PluginGroup and not ep_name_has_namespace(ep_name):
            msg = f"{epname_str}: Plugin name has no qualifying namespace!"
            raise ValueError(msg)

        name, version = from_ep_name(ep_name)
        p_ref = AnyPluginRef(group=self.name, name=name, version=version)

        if ep_name in self._ENTRY_POINTS:
            self._LOADED_PLUGINS.pop(p_ref, None)  # unload, if loaded
            pkg = ep_obj.dist
            msg = f"WARNING: {ep_name} is probably provided by multiple packages!\n"
            msg += f"The plugin will now be provided by: {pkg.name} {pkg.version}"
            eprint(msg)
        self._ENTRY_POINTS[ep_name] = ep_obj

        if ep_name not in self._VERSIONS:
            self._VERSIONS[name] = []
        self._VERSIONS[name].append(p_ref)
        self._VERSIONS[name].sort()  # should be cheap

    def __init__(self, entrypoints: Dict[str, EntryPoint]):
        self._ENTRY_POINTS = {}
        self._VERSIONS = {}

        for k, v in entrypoints.items():
            self._add_ep(k, v)

        self._LOADED_PLUGINS = {}
        self.__post_init__()

    def __post_init__(self):
        if type(self) is PluginGroup:
            # make the magic plugingroup plugin add itself for consistency
            ep_name = util.to_ep_name(self.Plugin.name, self.Plugin.version)
            ep_path = f"{type(self).__module__}:{type(self).__name__}"
            ep = EntryPoint(ep_name, ep_path, self.name)
            self._add_ep(ep_name, ep)

            self_ref = AnyPluginRef(
                group=self.name, name=self.name, version=self.Plugin.version
            )
            self._LOADED_PLUGINS[self_ref] = self
            self.provider(self_ref).plugins[self.name].append(self_ref)

    @property
    def name(self) -> str:
        """Return name of the plugin group."""
        return self.Plugin.name

    @property
    def packages(self) -> Dict[str, PluginPkgMeta]:
        """Return metadata of all packages providing metador plugins."""
        return dict(self._PKG_META)

    def versions(
        self, p_name: str, version: Optional[SemVerTuple] = None
    ) -> List[AnyPluginRef]:
        """Return installed versions of a plugin (compatible with given version)."""
        refs = list(self._VERSIONS.get(p_name) or [])
        if version is None:
            return refs
        requested = self.PluginRef(name=p_name, version=version)
        return [ref for ref in refs if ref.supports(requested)]

    def resolve(
        self, p_name: str, version: Optional[SemVerTuple] = None
    ) -> Optional[AnyPluginRef]:
        """Return most recent compatible version of a plugin."""
        if refs := self.versions(p_name, version):
            return refs[-1]  # latest (compatible) version
        return None

    def provider(self, ref: AnyPluginRef) -> PluginPkgMeta:
        """Return package metadata of Python package providing this plugin."""
        if type(self) is PluginGroup and ref.name == PG_GROUP_NAME:
            # special case - the mother plugingroup plugin is not an EP,
            # so we cheat a bit (schema is in same package, but is an EP)
            return self.provider(self.resolve("schema"))

        ep_name = util.to_ep_name(ref.name, ref.version)
        ep = self._ENTRY_POINTS[ep_name]
        return self._PKG_META[cast(Any, ep).dist.name]

    def is_plugin(self, p_cls):
        """Return whether this class is a (possibly marked) installed plugin."""
        if not isinstance(p_cls, type) or not issubclass(
            p_cls, self.Plugin.plugin_class
        ):
            return False

        c = UndefVersion._unwrap(p_cls) or p_cls  # get real underlying class
        # check its exactly a registered plugin, if it has a Plugin section
        if info := c.__dict__.get("Plugin"):
            if not isinstance(info, PluginBase):
                return False
            loaded_p = self._get_unsafe(info.name, info.version)
            return loaded_p is c
        else:
            return False

    # ----

    def __repr__(self):
        return f"<PluginGroup '{self.name}' {list(self.keys())}>"

    def __str__(self):
        def pg_line(name_refs):
            name, refs = name_refs
            vs = list(map(lambda x: to_semver_str(x.version), refs))
            # p = self.provider(pg_ref.name)
            # pkg = f"{p.name} {semver_str(p.version)}"
            return f"\t'{name}' ({', '.join(vs)})"

        pgs = "\n".join(map(pg_line, self._VERSIONS.items()))
        return f"Available '{self.name}' plugins:\n{pgs}"

    # ----
    # dict-like interface will provide latest versions of plugins by default

    def __contains__(self, key) -> bool:
        name, version = plugin_args(key)
        if pg_versions := self._VERSIONS.get(name):
            if not version:
                return True
            pg = self.PluginRef(name=name, version=version)
            return pg in pg_versions
        return False

    def __getitem__(self, key) -> Type[T]:
        if key not in self:
            raise KeyError(f"{self.name} not found: {key}")
        return self.get(key)

    def keys(self) -> Iterator[AnyPluginRef]:
        """Return all names of all plugins."""
        for pgs in self._VERSIONS.values():
            yield from pgs

    def values(self) -> Iterator[Type[T]]:
        """Return latest versions of all plugins (THIS LOADS ALL PLUGINS!)."""
        return map(self.__getitem__, self.keys())

    def items(self) -> Iterator[Tuple[AnyPluginRef, Type[T]]]:
        """Return pairs of plugin name and latest installed version (THIS LOADS ALL PLUGINS!)."""
        return map(lambda k: (k, self[k]), self.keys())

    # ----

    def _get_unsafe(self, p_name: str, version: Optional[SemVerTuple] = None):
        """Return most recent compatible version of given plugin name, without safety rails.

        Raises KeyError if no (compatible) schema found.

        For internal use only!
        """
        if ref := self.resolve(p_name, version):
            self._ensure_is_loaded(ref)
            return self._LOADED_PLUGINS[ref]
        else:  # error
            msg = f"{p_name}"
            if version:
                msg += f": no installed version is compatible with {version}"
            raise KeyError(msg)

    # inspired by this nice trick: https://stackoverflow.com/a/60362860
    PRX = TypeVar("PRX", bound="Type[T]")  # type: ignore

    @overload
    def get(self, key: str, version: Optional[SemVerTuple] = None) -> Optional[Type[T]]:
        ...  # pragma: no cover

    @overload
    def get(self, key: PRX, version: Optional[SemVerTuple] = None) -> Optional[PRX]:
        ...  # pragma: no cover

    def get(
        self, key: Union[str, PRX], version: Optional[SemVerTuple] = None
    ) -> Union[Type[T], PRX, None]:
        key_, version = plugin_args(key, version)

        # retrieve compatible plugin
        try:
            ret = self._get_unsafe(key_, version)
        except KeyError:
            return None

        if version is None:
            # no version constraint was passed or inferred -> mark it
            ret = UndefVersion._mark_class(ret)

        if isinstance(key, str):
            return cast(Type[T], ret)
        else:
            return ret

    # ----

    def _ensure_is_loaded(self, ref: AnyPluginRef):
        """Load plugin from entrypoint, if it is not loaded yet."""
        assert ref.group == self.name
        if ref in self._LOADED_PLUGINS:
            return  # already loaded, all good

        ep_name = util.to_ep_name(ref.name, ref.version)
        ret = self._ENTRY_POINTS[ep_name].load()
        self._LOADED_PLUGINS[ref] = ret

        self._load_plugin(ep_name, ret)

    def _explicit_plugin_deps(self, plugin) -> Set[AnyPluginRef]:
        """Return all plugin dependencies that must be taken into account."""
        def_deps = set(plugin.Plugin.requires)
        extra_deps = set(self.plugin_deps(plugin) or set())
        return def_deps.union(extra_deps)

    def plugin_deps(self, plugin) -> Set[AnyPluginRef]:
        """Return additional automatically inferred dependencies for a plugin."""

    def _load_plugin(self, ep_name: EPName, plugin):
        """Run checks and finalize loaded plugin."""
        from ..plugins import plugingroups

        # run inner Plugin class checks (with possibly new Fields cls)
        if not plugin.__dict__.get("Plugin"):
            raise TypeError(f"{ep_name}: {plugin} is missing Plugin inner class!")
        # pass ep_name to check that it agrees with the plugin info
        plugin.Plugin = self.Plugin.plugin_info_class.parse_info(
            plugin.Plugin, ep_name=ep_name
        )

        # do general checks first, if they fail no need to continue
        self._check_common(ep_name, plugin)
        self.check_plugin(ep_name, plugin)

        for dep_ref in self._explicit_plugin_deps(plugin):
            dep_grp = plugingroups[dep_ref.group]
            dep_grp._ensure_is_loaded(dep_ref)

        self.init_plugin(plugin)

    def _check_common(self, ep_name: EPName, plugin):
        """Perform both the common and specific checks a registered plugin.

        Raises a TypeError with message in case of failure.
        """
        # check correct base class of plugin, if stated
        if self.Plugin.plugin_class:
            util.check_is_subclass(ep_name, plugin, self.Plugin.plugin_class)

    def check_plugin(self, ep_name: EPName, plugin: Type[T]):
        """Perform plugin group specific checks on a registered plugin.

        Raises a TypeError with message in case of failure.

        To be overridden in subclasses for plugin group specific checks.

        Args:
            ep_name: Declared entrypoint name.
            plugin: Object the entrypoint is pointing to.
        """
        # NOTE: following cannot happen as long as we enforce
        # overriding check_plugin.
        # keep that here for now, in case we loosen this
        # if type(self) is not PluginGroup:
        #    return  # is not the "plugingroup" group itself

        # these are the checks done for other plugin group plugins:
        util.check_is_subclass(ep_name, plugin, PluginGroup)
        util.check_is_subclass(ep_name, self.Plugin.plugin_info_class, PluginBase)
        if plugin != PluginGroup:  # exclude itself. this IS its check_plugin
            util.check_implements_method(ep_name, plugin, PluginGroup.check_plugin)

        # NOTE: following cannot happen as long as we set the group
        # automatically using the metaclass.
        # keep that in case we decide to change that / get rid of the metaclass
        # ---
        # make sure that the declared plugin_info_class for the group sets 'group'
        # and it is also equal to the plugin group 'name'.
        # this is the safest way to make sure that Plugin.ref() works correctly.
        # ppgi_cls = plugin.Plugin.plugin_info_class
        # if not ppgi_cls.group:
        #    raise TypeError(f"{ep_name}: {ppgi_cls} is missing 'group' attribute!")
        # if not ppgi_cls.group == plugin.Plugin.name:
        #    msg = f"{ep_name}: {ppgi_cls.__name__}.group != {plugin.__name__}.Plugin.name!"
        #    raise TypeError(msg)

    def init_plugin(self, plugin: Type[T]):
        """Override this to do something after the plugin has been checked."""
        if type(self) is not PluginGroup:
            return  # is not the "plugingroup" group itself
        create_pg(plugin)  # create plugin group if it does not exist


# ----

_plugin_groups: Dict[str, PluginGroup] = {}
"""Instances of initialized plugin groups."""


def create_pg(pg_cls):
    """Create plugin group instance if it does not exist."""
    pg_ref = AnyPluginRef(
        group=PG_GROUP_NAME, name=pg_cls.Plugin.name, version=pg_cls.Plugin.version
    )
    if pg_ref in _plugin_groups:
        return _plugin_groups[pg_ref]

    if not isinstance(pg_cls.Plugin, PluginBase):
        # magic - substitute Plugin class with parsed plugin object
        pg_cls.Plugin = PGPlugin.parse_info(pg_cls.Plugin)

    # TODO: currently we cannot distinguish entrypoints
    # for different versions of the plugin group.
    # should not be problematic for now,
    # as the groups should not change much
    pg = pg_cls(get_group(pg_ref.name))
    _plugin_groups[pg_ref] = pg



"""Mkdocs hook to run tests with coverage collection and generate a badge."""
import logging
from io import StringIO
from pathlib import Path

import anybadge
import pytest
from coverage import Coverage

log = logging.getLogger("mkdocs")


badge_colors = {
    20: "red",
    40: "orange",
    60: "yellow",
    80: "greenyellow",
    90: "green",
}
"""Colors for overall coverage percentage (0-100)."""


def on_pre_build(config):
    """Generate coverage report if it is missing and create a badge."""
    if not Path("htmlcov").is_dir() or not Path(".coverage").is_file():
        log.info("Missing htmlcov or .coverage, running pytest to collect.")
        pytest.main(["--cov", "--cov-report=html"])
    else:
        log.info("Using existing coverage data.")

    cov = Coverage()
    cov.load()
    cov_percent = int(cov.report(file=StringIO()))
    log.info(f"Test Coverage: {cov_percent}%, generating badge.")

    badge = anybadge.Badge(
        "coverage",
        cov_percent,
        value_prefix=" ",
        value_suffix="% ",
        thresholds=badge_colors,
    )

    badge_svg = Path("docs/coverage_badge.svg")
    if badge_svg.is_file():
        badge_svg.unlink()
    badge.write_badge(badge_svg)



"""Immutable HDF5-based multi-container records.

This API **supports a subset of h5py**, namely reading, writing and deleting groups,
values and attributes. It **does not** support [hard, symbolic or external
links](https://docs.h5py.org/en/latest/high/group.html#link-classes),
so the data must be self-contained and strictly hierarchical.


**Correspondence of IH5 overlay classes and raw h5py classes:**

| IH5 API      | h5py         |
| ------------ | ------------ |
| `IH5Record`  | [h5py.File](https://docs.h5py.org/en/latest/high/file.html) |
| `IH5Group`   | [h5py.Group](https://docs.h5py.org/en/latest/high/group.html) |
| `IH5Dataset` | [h5py.Dataset](https://docs.h5py.org/en/latest/high/dataset.html) |
| `IH5AttributeManager` | [h5py.AttributeManager](https://docs.h5py.org/en/latest/high/attr.html) |

Anything that can be done to an IH5 Group, Dataset or AttributeManager can also be done to
the h5py counterparts.

If you are missing some functionality from h5py in the overlay classes,
please contact us or open an issue and we will see whether and how the missing
methods can be added in a coherent manner matching the IH5 semantics.

## Getting Started

A quite minimal working example:

```python
# just use IH5Record instead of h5py.File:
from metador_core.ih5.container import IH5Record


# initial creation:
with IH5Record("record_name", "w") as ds:
    # A new record is automatically in writable mode,
    # so let us write some data, just like with h5py:
    ds["foo/bar"] = "something"
    ds["foo"].attrs["attribute"] = 123

# updating the record we created later on:
with IH5Record("record_name", "r") as ds:
    # A record is opened in read-only mode, so
    # before we can add, modify or delete anything, we need to call:
    ds.create_patch()

    del ds["foo/bar"]
    ds["/foo/bar/baz"] = [1, 2, 3]
```

You SHOULD use `IH5Record` as a context manager, like in the example above.
If you for some reason do not, be aware that you MUST manually call `commit()`
to finilize your changes. You can also just use `close()`, which will also call
`commit()` for you. **Consider it good practice calling `commit()`
manually after completing your updates of the record.**
"""



from metador_core.container import MetadorContainer
from metador_core.harvester import file_harvester_pipeline, harvest, metadata_loader
from metador_core.packer.utils import embed_file


def test_pack(plugingroups_test, tutorialfiles, tmp_ds_path):
    """Pack a container with various file types."""
    schemas = plugingroups_test["schema"]
    harvesters = plugingroups_test["harvester"]

    BibMeta = schemas["core.bib"]
    DBMeta = schemas["core.dashboard"]
    ImgMeta = schemas["core.imagefile"]

    HrvFile = harvesters["core.file.generic"]
    HrvImgDim = harvesters["core.imagefile.dim"]
    ImgMetaLoader = metadata_loader(ImgMeta, use_sidecar=True)

    bibmeta_path = tutorialfiles("test.bibmeta.yaml")
    imgfile_path = tutorialfiles("test.png")
    jsonfile_path = tutorialfiles("test.json")
    mdfile_path = tutorialfiles("test.md")
    htmlfile_path = tutorialfiles("test.html")
    pdffile_path = tutorialfiles("test.pdf")

    image_pipeline = file_harvester_pipeline(HrvFile, HrvImgDim, ImgMetaLoader)
    imgmeta = harvest(ImgMeta, image_pipeline(imgfile_path))

    tmp_ds_path.mkdir()
    with MetadorContainer(tmp_ds_path / "test.h5", "w") as m:
        m.meta["core.bib"] = BibMeta.parse_file(bibmeta_path)

        node = embed_file(m, "foo/bar", imgfile_path, metadata=imgmeta)
        node.meta[DBMeta] = DBMeta.show(group=1)
        node = embed_file(m, "pdffile", pdffile_path)
        node.meta[DBMeta] = DBMeta.show(group=1, priority=10)
        node = embed_file(m, "jsonfile", jsonfile_path)
        node.meta[DBMeta] = DBMeta()
        node = embed_file(m, "mdfile", mdfile_path)
        node.meta[DBMeta] = DBMeta()
        node = embed_file(m, "htmlfile", htmlfile_path)
        node.meta[DBMeta] = DBMeta.show(
            [
                DBMeta.widget(widget_name="core.file.text.code", group=2),
                DBMeta.widget(group=2),
            ]
        )

    m = MetadorContainer("test.h5", "a")
    m.close()

    # TODO: assert some stuff



"""Generic container dashboard.

To **configure** a container dashboard: attach `DashboardConf` metadata to `MetadorContainer` nodes.

To **show** a container dashboard: create a `Dashboard` instance.
"""
from __future__ import annotations

from functools import partial
from itertools import groupby
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Tuple

import panel as pn
from panel.viewable import Viewable
from phantom.interval import Inclusive

from ..container import MetadorContainer, MetadorNode
from ..plugins import schemas, widgets
from ..schema import MetadataSchema
from ..schema.plugins import PluginRef
from ..schema.types import NonEmptyStr, SemVerTuple

if TYPE_CHECKING:
    from . import WidgetServer


class DashboardPriority(int, Inclusive, low=1, high=10):
    """Dashboard priority of a widget."""


class DashboardGroup(int, Inclusive, low=1):
    """Dashboard group of a widget."""


class WidgetConf(MetadataSchema):
    """Configuration of a widget in the dashboard."""

    priority: Optional[DashboardPriority] = DashboardPriority(1)
    """Priority of the widget (1-10), higher priority nodes are shown first."""

    group: Optional[DashboardGroup]
    """Dashboard group of the widget.

    Groups are presented in ascending order.
    Widgets are ordered by priority within a group.
    All widgets in a group are shown in a single row.

    Widgets without an assigned group come last.
    """

    # ----

    schema_name: Optional[NonEmptyStr]
    """Name of schema of an metadata object at the current node that is to be visualized.

    If not given, any suitable will be selected if possible.
    """

    schema_version: Optional[SemVerTuple]
    """Version of schema to be used.

    If not given, any suitable will be selected if possible.
    """

    widget_name: Optional[str]
    """Name of widget to be used.

    If not given, any suitable will be selected if possible.
    """

    widget_version: Optional[SemVerTuple]
    """Version of widget to be used.

    If not given, any suitable will be selected if possible.
    """


class DashboardConf(MetadataSchema):
    """Schema describing dashboard configuration for a node in a container.

    Instantiating without passing a list of widget configurations will
    return an instance that will show an arbitrary suitable widget, i.e.
    is equivalent to `DashboardConf.show()`
    """

    class Plugin:
        name = "core.dashboard"
        version = (0, 1, 0)

    widgets: List[WidgetConf] = [WidgetConf()]
    """Widgets to present for this node in the dashboard.

    If left empty, will try present any widget usable for this node.
    """

    @staticmethod
    def widget(**kwargs) -> WidgetConf:
        """Construct a dashboard widget configuration (see `WidgetConf`)."""
        # for convenience
        return WidgetConf(**kwargs)

    @classmethod
    def show(cls, _arg: List[WidgetConf] = None, **kwargs):
        """Construct a dashboard configuration for the widget(s) of one container node.

        For one widget, pass the widget config (if any) as keyword arguments,
        e.g.  `DashboardConf.show(group=1)`.

        For multiple widgets, create widget configurations with `widget(...)`,
        and pass them to `show`, e.g.:
        `DashboardConf.show([DashboardConf.widget(), DashboardConf.widget(group=2)])`.
        """
        if _arg and kwargs:
            msg = "Pass widget config arguments or list of widget configs - not both!"
            raise ValueError(msg)

        if _arg is None:
            # kwargs have config for a singleton widget
            widgets = [cls.widget(**kwargs)]
        else:
            # multiple widgets, preconfigured
            widgets = list(_arg)
        return cls(widgets=widgets)


# ----

NodeWidgetPair = Tuple[MetadorNode, WidgetConf]
"""A container node paired up with a widget configuration."""

NodeWidgetRow = List[NodeWidgetPair]
"""Sorted list of NodeWidgetPairs.

Ordered first by descending priority, then by ascending node path.
"""


def sorted_widgets(
    widgets: Iterable[NodeWidgetPair],
) -> Tuple[Dict[int, NodeWidgetRow], NodeWidgetRow]:
    """Return widgets in groups, ordered by priority and node path.

    Returns tuple with dict of groups and a remainder of ungrouped widgets.
    """

    def nwp_group(tup: NodeWidgetPair) -> int:
        return tup[1].group or 0

    def nwp_prio(tup: NodeWidgetPair) -> int:
        return -tup[1].priority or 0  # in descending order of priority

    def sorted_group(ws: Iterable[NodeWidgetPair]) -> NodeWidgetRow:
        """Sort first on priority, and for same priority on container node."""
        return list(sorted(sorted(ws, key=lambda x: x[0].name), key=nwp_prio))

    # dict, sorted in ascending group order (but ungrouped are 0)
    ret = dict(
        sorted(
            {
                k: sorted_group(v)
                for k, v in groupby(sorted(widgets, key=nwp_group), key=nwp_group)
            }.items()
        )
    )
    ungrp = ret.pop(0, [])  # separate out the ungrouped (were mapped to 0)
    return ret, ungrp


# ----


def _resolve_schema(node: MetadorNode, wmeta: WidgetConf) -> PluginRef:
    """Return usable schema+version pair for the node based on widget metadata.

    If a schema name or schema version is missing, will complete these values.

    Usable schema means that:
    * there exists a compatible installed schema
    * there exists a compatible metadata object at given node

    Raises ValueError on failure to find a suitable schema.
    """
    if wmeta.schema_name is None:
        # if no schema selected -> pick any schema for which we have:
        #   * a schema instance at the current node
        #   * installed widget(s) that support it
        for obj_schema in node.meta.query():
            if next(widgets.widgets_for(obj_schema), None):
                return obj_schema

        msg = f"Cannot find suitable schema for a widget at node: {node.name}"
        raise ValueError(msg)

    # check that a node object is compatible with the one requested
    req_ver = wmeta.schema_version if wmeta.schema_version else "any"
    req_schema = f"{wmeta.schema_name} ({req_ver})"
    s_ref = next(node.meta.query(wmeta.schema_name, wmeta.schema_version), None)
    if s_ref is None:
        msg = f"Dashboard wants metadata compatible with {req_schema}, but node"
        if nrf := next(node.meta.query(wmeta.schema_name), None):
            nobj_schema = f"{nrf.name} {nrf.version}"
            msg += f"only has incompatible object: {nobj_schema}"
        else:
            msg += "has no suitable object"
        raise ValueError(msg)

    # if no version is specified, pick the one actually present at the node
    version = wmeta.schema_version or s_ref.version
    s_ref = schemas.PluginRef(name=wmeta.schema_name, version=version)

    # ensure there is an installed schema compatible with the one requested
    # (NOTE: if child schemas exist, the parents do too - no need to check)
    installed_schema = schemas.get(s_ref.name, s_ref.version)
    if installed_schema is None:
        msg = f"No installed schema is compatible with {req_schema}"
        raise ValueError(msg)

    return s_ref


def _widget_suitable_for(m_obj: MetadataSchema, w_ref: PluginRef) -> bool:
    """Granular check whether the widget actually works with the metadata object.

    Assumes that the passed object is known to be one of the supported schemas.
    """
    w_cls = widgets._get_unsafe(w_ref.name, w_ref.version)
    return w_cls.Plugin.primary and w_cls.supports_meta(m_obj)


def _resolve_widget(
    node: MetadorNode,
    s_ref: PluginRef,
    w_name: Optional[str],
    w_version: Optional[SemVerTuple],
) -> PluginRef:
    """Return suitable widget for the node based on given dashboard metadata."""
    if w_name is None:
        # get candidate widgets in alphabetic order (all that claim to work with schema)
        cand_widgets = sorted(widgets.widgets_for(s_ref))
        is_suitable = partial(_widget_suitable_for, node.meta[s_ref.name])
        # filter out the ones that ACTUALLY can handle the object and are eligible
        if w_ref := next(filter(is_suitable, cand_widgets), None):
            return w_ref
        else:
            msg = f"Could not find suitable widget for {w_name} at node {node.name}"
            raise ValueError(msg)

    # now we have a widget name (and possibly version) - check it
    widget_class = widgets._get_unsafe(w_name, w_version)
    if widget_class is None:
        raise ValueError(f"Could not find compatible widget: {w_name} {w_version}")
    if not widget_class.supports(*schemas.parent_path(s_ref.name, s_ref.version)):
        msg = f"Widget {widget_class.Plugin.ref()} does not support {s_ref}"
        raise ValueError(msg)

    w_ref = widget_class.Plugin.ref()
    return w_ref


# ----


def get_grp_label(idx):
    """Create and return a styled group label."""
    return pn.pane.Str(
        f"Group {idx+1}" if idx is not None else "Ungrouped resources",
        style={
            "font-size": "15px",
            "font-weight": "bold",
            "text-decoration": "underline",
        },
    )


def add_widgets(w_grp, ui_grp, *, server=None, container_id: Optional[str] = None):
    """Instantiate and add widget to the flexibly wrapping row that handles the entire group."""
    w_width, w_height = 500, 500  # max size of a widget tile, arbitrarily set
    for node, wmeta in w_grp:
        w_cls = widgets.get(wmeta.widget_name, wmeta.widget_version)
        label = pn.pane.Str(f"{node.name}:")

        # instantiating the appropriate widget
        w_obj = w_cls(
            node,
            wmeta.schema_name,
            wmeta.schema_version,
            server=server,
            container_id=container_id,
            # reset max widget of a widget tile,  only if it is for a pdf, text or video file
            max_width=700
            if "pdf" in wmeta.widget_name
            or "text" in wmeta.widget_name
            or "video" in wmeta.widget_name
            else w_width,
            # reset max height of a widget tile, only if it is for a text file
            max_height=700 if "text" in wmeta.widget_name else w_height,
        )

        # adding the new widget to the given row
        ui_grp.append(
            pn.Column(
                label,
                w_obj.show(),
                sizing_mode="scale_both",
                scroll=False
                if "image" in wmeta.widget_name or "pdf" in wmeta.widget_name
                else True,
            )
        )
    return ui_grp


def get_grp_row(
    *,
    idx=None,
    widget_group=None,
    divider=False,
    server=None,
    container_id: Optional[str] = None,
):
    """Create a flexible and wrapping row for all widgets within a single group."""
    return pn.FlexBox(
        get_grp_label(idx=idx),
        add_widgets(
            widget_group,
            pn.FlexBox(
                flex_direction="row",
                justify_content="space-evenly",
                align_content="space-evenly",
                align_items="center",
                sizing_mode="scale_both",
            ),
            server=server,
            container_id=container_id,
        ),
        pn.layout.Divider(margin=(100, 0, 20, 0)) if divider == True else None,
        flex_direction="column",
        justify_content="space-evenly",
        align_content="space-evenly",
        align_items="center",
        sizing_mode="scale_both",
    )


class Dashboard:
    """The dashboard presents a view of all marked nodes in a container.

    To be included in the dashboard, a node must be marked by a `DashboardConf`
    object configuring at least one widget for that node.


    Note that the `Dashboard` needs
    * either a widget server to be passed (embedding in a website),
    * or the container is wrapped by `metador_core.widget.jupyter.Previewable` (notebook mode)
    """

    def __init__(
        self,
        container: MetadorContainer,
        *,
        server: WidgetServer = None,
        container_id: Optional[str] = None,
    ):
        """Return instance of a dashboard.

        Args:
            container: Actual Metador container that is open and readable
            server: `WidgetServer` to use for the widgets (default: standalone server / Jupyter mode)
            container_id: Container id usable with the server to get this container (default: container UUID)
        """
        self._container: MetadorContainer = container
        self._server = server
        self._container_id: str = container_id

        # figure out what schemas to show and what widgets to use and collect
        ws: List[NodeWidgetPair] = []
        for node in self._container.metador.query(DashboardConf):
            dbmeta = node.meta.get(DashboardConf)
            restr_node = node.restrict(read_only=True, local_only=True)
            for wmeta in dbmeta.widgets:
                ws.append((restr_node, self._resolve_node(node, wmeta)))

        grps, ungrp = sorted_widgets(ws)
        self._groups = grps
        self._ungrouped = ungrp

    def _resolve_node(self, node: MetadorNode, wmeta: WidgetConf) -> WidgetConf:
        """Check and resolve widget dashboard metadata for a node."""
        wmeta = wmeta.copy()  # use copy, abandon original

        s_ref: PluginRef = _resolve_schema(node, wmeta)
        wmeta.schema_name = s_ref.name
        wmeta.schema_version = s_ref.version

        w_ref: PluginRef = _resolve_widget(
            node, s_ref, wmeta.widget_name, wmeta.widget_version
        )
        wmeta.widget_name = w_ref.name
        wmeta.widget_version = w_ref.version

        return wmeta

    def show(self) -> Viewable:
        """Instantiate widgets for container and return resulting dashboard."""
        # Outermost element: The Dashboard is a column of widget groups
        db = pn.FlexBox(
            flex_direction="column",
            justify_content="space-evenly",
            align_content="space-evenly",
            align_items="center",
            sizing_mode="scale_both",
        )

        # add each widget group within individual, flexibly-wrapping rows
        for idx, widget_group in enumerate(self._groups.values()):
            db.append(
                get_grp_row(
                    idx=idx,
                    widget_group=widget_group,
                    divider=True,
                    server=self._server,
                    container_id=self._container_id,
                )
            )

        # dump remaining ungrouped widgets into a separate flexibly-wrapping row
        ungrp_exist = len(self._ungrouped) != 0
        if ungrp_exist:
            db.append(
                get_grp_row(
                    widget_group=self._ungrouped,
                    server=self._server,
                    container_id=self._container_id,
                )
            )
        return db



from __future__ import annotations

import hashlib
import os
from io import BytesIO
from pathlib import Path
from typing import Any, BinaryIO, Dict, Optional, Union

_hash_alg = {
    # "md5": hashlib.md5,
    # "sha1": hashlib.sha1,
    "sha256": hashlib.sha256,
    "sha512": hashlib.sha512,
}
"""Supported hashsum algorithms."""


def hashsum(data: Union[bytes, BinaryIO], alg: str):
    """Compute hashsum from given binary file stream using selected algorithm."""
    if isinstance(data, bytes):
        data = BytesIO(data)
    try:
        h = _hash_alg[alg]()
    except KeyError:
        raise ValueError(f"Unsupported hashsum: {alg}")

    while True:
        chunk = data.read(h.block_size)
        if not chunk:
            break
        h.update(chunk)

    return h.hexdigest()


DEF_HASH_ALG = "sha256"
"""Algorithm to use and string to prepend to a resulting hashsum."""


def qualified_hashsum(data: Union[bytes, BinaryIO], alg: str = DEF_HASH_ALG):
    """Like hashsum, but prepends the algorithm to the string."""
    return f"{alg}:{hashsum(data, alg)}"


def file_hashsum(path: Path, alg: str = DEF_HASH_ALG):
    with open(path, "rb") as f:
        return qualified_hashsum(f, alg)


DirHashsums = Dict[str, Any]
"""Nested dict representing a directory.

str values represent files (checksum) or symlinks (target path),
dict values represent sub-directories.
"""


def rel_symlink(base: Path, dir: Path) -> Optional[Path]:
    """From base path and a symlink path, normalize it to be relative to base.

    Mainly used to eliminate .. in paths.

    If path points outside base, returns None.
    """
    path = dir.parent / os.readlink(str(dir))
    try:
        return path.resolve().relative_to(base.resolve())
    except ValueError:
        return None  # link points outside of base directory


def dir_hashsums(dir: Path, alg: str = DEF_HASH_ALG) -> DirHashsums:
    """Return hashsums of all files.

    Resulting paths are relative to the provided `dir`.

    In-directory symlinks are treated like files and the target is stored
    instead of computing a checksum.

    Out-of-directory symlinks are not allowed.
    """
    ret: Dict[str, Any] = {}
    for path in dir.rglob("*"):
        is_file, is_sym = path.is_file(), path.is_symlink()
        relpath = path.relative_to(dir)

        fname = None
        val = ""

        if is_file or is_sym:
            fname = relpath.name
            relpath = relpath.parent  # directory dicts to create = up to parent

        if is_file:
            val = file_hashsum(path, alg)  # value = hashsum
        elif is_sym:
            sym_trg = rel_symlink(dir, path)
            if sym_trg is None:
                raise ValueError(f"Symlink inside '{dir}' points to the outside!")
            val = "symlink:" + str(sym_trg)  # value = symlink target

        # create nested dicts, if not existing yet
        curr = ret
        for seg in str(relpath).split("/"):
            if seg == ".":
                continue
            if seg not in curr:
                curr[seg] = dict()
            curr = curr[seg]
        # store file hashsum or symlink target
        if is_file or is_sym:
            assert fname is not None
            curr[fname] = val

    return ret


# ----



"""Test dirdiff structure."""
from pathlib import Path

import pytest

from metador_core.util.diff import DiffNode, DirDiff
from metador_core.util.hashsums import dir_hashsums

pytest.skip(reason="invenio installation problem", allow_module_level=True)


def test_dir_hashsums_invalid_symlink_fail(tmp_ds_path):
    """Directories with symlinks pointing out are forbidden."""
    tmp_ds_path.mkdir()
    outside = tmp_ds_path.resolve().parent
    Path(tmp_ds_path / "sym").symlink_to(outside)

    with pytest.raises(ValueError):
        dir_hashsums(tmp_ds_path, "sha256")


def test_dirdiff_same_empty(testinputs):
    """Comparing a directory with itself should produce empty diff."""
    ds = testinputs("dirdiff1")
    dhs = dir_hashsums(ds, "sha256")
    diff = DirDiff.compare(dhs, dhs)
    assert diff.is_empty


def test_dir_hashsums(testinputs):
    ds = testinputs("dirdiff1")
    expected = {
        "_meta.yaml": "sha256:5f5480ca66e7014685c6eff80f4b6535419746517328c0a5d2f839b1a7d58de7",
        "a": {
            "b": {
                "d": "symlink:a/b",
                "c.csv": "sha256:d37ad046c2e0c067e59f64a91732873f31df478d51a26e3552e2522cade609b2",
                "c.csv_meta.yaml": "sha256:833c8d30f975d23cd0250900a5b3efa26f04a45d14c0c492a1172f9a41af9146",
            }
        },
        "e": "sha256:8fa6d6dee361e59354df82c22ba29f9ae0d45f6560a5d906b156f759b25f1b21",
        "d": "symlink:a/b",
        "f": "sha256:a0742cb9bf98c1adec87020f4b3d5a0b8db4c6c48e2bce72aaa4b77604ce948f",
        "z": "sha256:e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
    }
    assert dir_hashsums(ds, "sha256") == expected
    assert dir_hashsums(ds) == expected  # default algorithm is sha256


def test_dirdiff_nontrivial(testinputs):
    ds1, ds2 = testinputs("dirdiff1"), testinputs("dirdiff2")
    dhs1, dhs2 = dir_hashsums(ds1, "sha256"), dir_hashsums(ds2, "sha256")

    diff = DirDiff.compare(dhs1, dhs2)
    assert not diff.is_empty

    # test DirDiff.get()
    with pytest.raises(AssertionError):
        diff.get(Path("/a/b"))
    assert DirDiff.compare(dhs2, dhs2).get("a") is None  # type: ignore
    assert diff.get(Path("a/x")) is None
    ab = diff.get(Path("a/b"))
    assert ab is not None

    assert diff.status(ab) == DiffNode.Status.modified
    assert ab.prev_type == DiffNode.ObjType.directory
    assert ab.curr_type == DiffNode.ObjType.file

    d = diff.get(Path("d"))
    assert d.status() == DiffNode.Status.removed
    assert d.prev_type == DiffNode.ObjType.symlink
    assert d.curr_type is None

    e = diff.get(Path("e"))
    assert e.status() == DiffNode.Status.modified
    assert e.prev_type == DiffNode.ObjType.file
    assert e.curr_type == DiffNode.ObjType.directory

    eg = diff.get(Path("e/g"))
    assert eg.status() == DiffNode.Status.added
    assert eg.prev_type is None
    assert eg.curr_type == DiffNode.ObjType.file

    f = diff.get(Path("f"))
    assert f.status() == DiffNode.Status.modified
    assert f.prev_type == DiffNode.ObjType.file
    assert f.curr_type == DiffNode.ObjType.file

    assert diff.get(Path("_meta.yaml")) is None
    assert diff.status(diff.get(Path("_meta.yaml"))) == DiffNode.Status.unchanged

    # test annotate()

    # changed - check correct order rec(removed, modified, self, added) + unchanged
    lst = ["d", "a/b/c.csv", "a/b/c.csv_meta.yaml", "a/b/d", "a/b", "a"]
    lst += ["e", "e/g", "f", ".", "h", "_meta.yaml", "z"]
    exp_paths = [ds2 / x for x in lst]

    ordered_paths = list(diff.annotate(ds2).keys())
    assert ordered_paths == exp_paths

    # unchanged -> empty annotate
    assert DirDiff.compare(dhs2, dhs2).annotate(ds2) == {}  # type: ignore



"""Various helper functions that packers can use."""

from __future__ import annotations

import urllib.parse
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Optional, Union

import h5py
import numpy
from pydantic import ValidationError

from ..container import MetadorContainer, MetadorDataset, MetadorGroup
from ..harvester import harvest
from ..plugins import harvesters, schemas
from ..schema import MetadataSchema
from .types import DirValidationErrors


def _h5_wrap_bytes(bs: bytes):
    """Wrap bytes with numpy.void if non-empty, or return h5py.Empty."""
    # NOTE: need to think what to do in case we provide non-HDF backends
    # should not be provided directly to users! it's an internal detail!
    return numpy.void(bs) if len(bs) else h5py.Empty("b")


def check_metadata_file(path: Path, **kwargs):
    """Check a metadata file, return error object.

    If `required` is set, will add an error if file is missing.

    If `schema` is passed and file exists, will validate the file and log errors.

    Combine both to check that a file does exist and is valid according to a schema.
    """
    required: bool = kwargs.get("required", False)
    schema: MetadataSchema = kwargs.get("schema", None)
    errs = DirValidationErrors()

    exists = path.is_file()
    if required and not exists:
        errs.add(str(path), f"Required metadata file not found: '{path}'")
    if schema is not None and exists:
        try:
            schema.parse_file(path)
        except (JSONDecodeError, ValidationError, FileNotFoundError) as e:
            errs.add(str(path), str(e))

    return errs


FileMeta = schemas.get("core.file", (0, 1, 0))


def embed_file(
    node: Union[MetadorContainer, MetadorGroup],
    node_path: str,
    file_path: Union[Path, str],
    *,
    metadata: Optional[MetadataSchema] = None,
) -> MetadorDataset:
    """Embed a file, adding minimal generic metadata to it.

    Will also ensure that the attached metadata has RO-Crate compatible @id set correctly.

    Args:
        node: Container where to embed the file contents
        node_path: Fresh path in container where to place the file
        file_path: Path of an existing file to be embedded
        metadata: If provided, will attach this instead of harvesting defaults.

    Returns:
        Dataset of new embedded file.
    """
    file_path = Path(file_path)

    # check container and file
    if node_path in node:
        raise ValueError(f"Path '{node}' already exists in given container or group!")
    if not file_path.is_file():
        raise ValueError(f"Path '{file_path}' does not look like an existing file!")

    if not metadata:
        # no metadata given -> harvest minimal information about a file
        hv_file = harvesters["core.file.generic"]
        metadata = harvest(FileMeta, [hv_file(filepath=file_path)])
    else:
        metadata = metadata.copy()  # defensive copying!

    # check metadata
    if not isinstance(metadata, FileMeta):
        msg = f"Provided metadata is not compatible with '{FileMeta.Plugin.name}'!"
        raise ValueError(msg)
    if not schemas.is_plugin(type(metadata)):
        msg = f"Given metadata is a {type(metadata)}, which is not a schema plugin!"
        raise ValueError(msg)

    data = _h5_wrap_bytes(file_path.read_bytes())
    ret = node.create_dataset(node_path, data=data)

    # set file metadata @id to be relative to dataset root just like RO Crate wants
    metadata.id_ = urllib.parse.quote(f".{ret.name}")

    # embed file and metadata in container:
    ret.meta[metadata.Plugin.name] = metadata
    return ret



import magic
from PIL import Image

from ..harvester import FileHarvester
from ..plugins import schemas
from ..util.hashsums import hashsum

FileMeta = schemas["core.file"]
ImageFileMeta = schemas["core.imagefile"]
BibMeta = schemas["core.bib"]
TableMeta = schemas["core.table"]


class FileMetaHarvester(FileHarvester):
    """Default harvester for basic common.file metadata.

    Harvests file name, file size, mimetype and hashsum of the file.
    """

    class Plugin:
        name = "core.file.generic"
        version = (0, 1, 0)
        returns = schemas.PluginRef(name="core.file", version=(0, 1, 0))

    def run(self):
        path = self.args.filepath

        sz = path.stat().st_size
        hs = hashsum(open(path, "rb"), "sha256")
        mt = magic.from_file(path, mime=True)
        return self.schema(
            filename=path.name, contentSize=sz, sha256=hs, encodingFormat=mt
        )


class ImageFileMetaHarvester(FileHarvester):
    """Harvester to obtain dimensions (width and height) of an image file."""

    class Plugin:
        name = "core.imagefile.dim"
        version = (0, 1, 0)
        returns = schemas.PluginRef(name="core.imagefile", version=(0, 1, 0))

    def run(self):
        path = self.args.filepath

        with Image.open(path) as img:
            width, height = img.size
        return self.schema(width=width, height=height)



import secrets
import shutil
from pathlib import Path

import pytest


@pytest.fixture(scope="module")
def plugingroups_test():
    """Access to plugingroups in a test, but will reset afterwards."""
    from metador_core.plugins import plugingroups

    yield plugingroups

    plugingroups.__reset__()


@pytest.fixture(scope="module")
def schemas():
    """Access available schemas in a test without needing to import.

    Could be more robust in case the plugin system breaks.
    """
    from metador_core.plugins import schemas

    return schemas


# ----


@pytest.fixture(scope="session")
def ds_dir(tmpdir_factory):
    """Create a fresh temporary directory for records created in the tests."""
    return tmpdir_factory.mktemp("metador_tests")


@pytest.fixture
def tmp_ds_path_factory(ds_dir):
    """Return a record name generator to be used for creating records.

    All containers of all records will be cleaned up after completing the test.
    """
    names = []

    def fresh_name() -> Path:
        name = secrets.token_hex(4)
        names.append(name)
        path = Path(ds_dir / name)
        return path

    yield fresh_name

    # clean up
    for name in names:
        for path in Path(ds_dir).glob(f"{name}*"):
            if path.is_file() or path.is_symlink():
                path.unlink()
            elif path.is_dir():
                shutil.rmtree(path)


@pytest.fixture
def tmp_ds_path(tmp_ds_path_factory):
    """Generate a record name to be used for creating records.

    All containers will be cleaned up after completing the test.
    """
    return tmp_ds_path_factory()


@pytest.fixture
def testinput_maker(tmp_ds_path_factory):
    def wrapped(root_dir: Path):
        def copy_from(name: str):
            dst_dir = tmp_ds_path_factory()
            dst_dir.mkdir()
            dst = dst_dir / name
            assert not dst.exists()

            src = root_dir / name
            if src.is_file():
                shutil.copy2(src, dst)
            elif src.is_dir():
                shutil.copytree(src, dst, symlinks=True)
            else:
                raise RuntimeError(f"No such test input in {root_dir}: {name}")

            assert dst.exists()
            return dst

        return copy_from

    return wrapped


TEST_DIR = Path(__file__).resolve().parent

TEST_DATA_DIR = TEST_DIR / "data"
"""Location of the test input data."""

TUTORIAL_FILES_DIR = TEST_DIR.parent / "docs" / "notebooks" / "files"
"""Location of the tutorial notebook input files."""


@pytest.fixture
def testinputs(testinput_maker):
    """Create temporary file or directory based on a known test input."""
    return testinput_maker(TEST_DATA_DIR)


@pytest.fixture
def tutorialfiles(testinput_maker):
    """Create temporary file or directory based on a known tutorial input."""
    return testinput_maker(TUTORIAL_FILES_DIR)



from typing import Dict, Iterator, Set, Type

from pydantic import BaseModel
from pydantic.fields import ModelField

from .typing import get_annotations, get_type_hints, is_subclass_of, traverse_typehint


def field_origins(m: Type[BaseModel], name: str) -> Iterator[Type[BaseModel]]:
    """Return sequence of bases where the field type hint was defined / overridden."""
    return (
        b for b in m.__mro__ if issubclass(b, BaseModel) and name in get_annotations(b)
    )


def updated_fields(m: Type[BaseModel]) -> Set[str]:
    """Return subset of fields that are added or overridden by a new type hint."""
    return {n for n in m.__fields__.keys() if next(field_origins(m, n)) is m}


def new_fields(m: Type[BaseModel]) -> Set[str]:
    # return {n for n in updated_fields(m) if not next(field_origins(m.__base__, n), None)}
    return set(m.__fields__.keys()) - set(m.__base__.__fields__.keys())  # type: ignore


def field_atomic_types(mf: ModelField, *, bound=object) -> Iterator[Type]:
    """Return sequence of nested atomic types in the hint of given field."""
    return filter(is_subclass_of(bound), traverse_typehint(mf.type_))


def atomic_types(m: BaseModel, *, bound=object) -> Dict[str, Set[Type]]:
    """Return dict from field name to model classes referenced in the field definition.

    Args:
        m: Pydantic model
        bound (object): If provided, will be used to filter results to
            contain only subclasses of the bound.
    """
    return {k: set(field_atomic_types(v, bound=bound)) for k, v in m.__fields__.items()}


def field_parent_type(m: Type[BaseModel], name: str) -> Type[BaseModel]:
    """Return type of field assigned in the next parent that provides a type hint."""
    b = next(filter(lambda x: x is not m, field_origins(m, name)), None)
    if not b:
        raise ValueError(f"No base class of {m} defines a field called '{name}'!")
    return get_type_hints(b).get(name)



import pytest
from hypothesis import assume, given
from hypothesis import strategies as st
from pydantic import ValidationError

FILE_DICT = dict(
    filename="test.txt", encodingFormat="text/plain", contentSize=0, sha256="abc"
)


@given(st.sets(st.sampled_from(list(FILE_DICT.keys()))))
def test_filemeta(schemas_test, ks):
    assume(len(ks) < len(FILE_DICT))

    FileMeta = schemas_test["core.file"]
    with pytest.raises(ValidationError):
        FileMeta(**{k: v for k, v in FILE_DICT.items() if k in ks})


def test_filemeta2(schemas_test):
    FileMeta = schemas_test["core.file"]
    m = FileMeta(**FILE_DICT)
    assert m.dict().get("@context").find("ro/crate") > 0
    assert m.dict().get("@type") == "File"


DIR_DICT = dict(name="directory")


def test_dirmeta(schemas_test):
    DirMeta = schemas_test["core.dir"]
    m = DirMeta(**DIR_DICT)
    assert m.dict().get("@context").find("ro/crate") > 0
    assert m.dict().get("@type") == "Dataset"


# TODO: test organization
BIB_DICT = dict(
    name="My dataset",
    abstract="some description",
    dateCreated="2123-04-05",
    creator=dict(name="Jane Doe"),
    author=[],
)


def test_bibmeta(schemas_test):
    BibMeta = schemas_test["core.bib"]
    m = BibMeta(**BIB_DICT)
    print(m)
    assert m.dict().get("@context").find("ro/crate") > 0
    assert m.dict().get("@type") == "Dataset"


# TODO: test imagefile meta, TableMeta, SIValue, NumValue



import pytest

from metador_core.plugin.metaclass import PluginMetaclassMixin, UndefVersion
from metador_core.plugin.types import to_semver_str


class SomeBasePlugin(metaclass=PluginMetaclassMixin):
    pass


class PluginA(SomeBasePlugin):
    class Plugin:
        name = "plugin_a"
        version = (0, 1, 0)


def test_marker_mixin():
    MarkedA = UndefVersion._mark_class(PluginA)
    with pytest.raises(TypeError):
        UndefVersion._mark_class(MarkedA)  # already marked

    # looks like original plugin wrt. to plugin info
    assert MarkedA.Plugin == PluginA.Plugin

    assert not UndefVersion._is_marked(PluginA)
    assert UndefVersion._is_marked(MarkedA)

    assert UndefVersion._unwrap(MarkedA) is PluginA
    assert UndefVersion._unwrap(PluginA) is None


def test_subclass_plugin():
    MarkedA = UndefVersion._mark_class(PluginA)
    MarkedB = UndefVersion._mark_class(SomeBasePlugin)

    # inherit from plugin -> should work
    class PluginC(PluginA):
        ...

    # but must not inherit Plugin info class
    assert PluginC.Plugin is None

    # must not inherit from marked class...
    with pytest.raises(TypeError):  # with inner Plugin

        class PluginD(MarkedA):
            ...

    with pytest.raises(TypeError):  # without inner Plugin

        class PluginE(MarkedB):
            ...


def test_metaclass_repr():
    # repr of a plugin should show name and version
    repr_str = repr(PluginA)
    assert repr_str.find(PluginA.Plugin.name)
    assert repr_str.find(to_semver_str(PluginA.Plugin.version))

    # repr of marked plugin should show that the version was chosen impicitly
    MarkedA = UndefVersion._mark_class(PluginA)
    assert repr(MarkedA).find("unspecified")



import json
from pathlib import Path
from typing import Union

from pydantic import BaseModel, Extra, ValidationError
from pydantic_yaml import parse_yaml_file_as, parse_yaml_raw_as, to_yaml_str

from .encoder import DynEncoderModelMetaclass
from .parser import ParserMixin


def _mod_def_dump_args(kwargs):
    """Set `by_alias=True` in given kwargs dict, if not set explicitly."""
    if "by_alias" not in kwargs:
        kwargs["by_alias"] = True  # e.g. so we get correct @id, etc fields
    if "exclude_none" not in kwargs:
        kwargs["exclude_none"] = True  # we treat None as "missing" so leave it out
    return kwargs


class BaseModelPlus(ParserMixin, BaseModel, metaclass=DynEncoderModelMetaclass):
    """Extended pydantic BaseModel with some good defaults.

    Used as basis for various entities, including:
    * Metadata schemas
    * Harvester arguments
    """

    class Config:
        # keep extra fields by default
        extra = Extra.allow
        # make PrivateAttr wrappers not always needed
        underscore_attrs_are_private = True
        # serialize enums properly
        use_enum_values = True
        # when alias is set, still allow using field name
        # (we use aliases for invalid attribute names in Python)
        allow_population_by_field_name = True
        # users should jump through hoops to add invalid stuff
        validate_assignment = True
        # defaults should also be validated
        validate_all = True
        # for JSON compat
        allow_inf_nan = False
        # pydantic anystr config: non-empty, non-whitespace
        # (but we prefer NonEmptyStr anyway for inheritance)
        anystr_strip_whitespace = True
        min_anystr_length = 1

    def dict(self, *args, **kwargs):
        """Return a dict.

        Nota that this will eliminate all pydantic models,
        but might still contain complex value types.
        """
        return super().dict(*args, **_mod_def_dump_args(kwargs))

    def json(self, *args, **kwargs) -> str:
        """Return serialized JSON as string."""
        return super().json(*args, **_mod_def_dump_args(kwargs))

    def json_dict(self, **kwargs):
        """Return a JSON-compatible dict.

        Uses round-trip through JSON serialization.
        """
        return json.loads(self.json(**kwargs))

    def yaml(self, **kwargs) -> str:
        """Return serialized YAML as string."""
        # Current way: use round trip through JSON to kick out non-JSON entities
        # (more elegant: allow ruamel yaml to reuse defined custom JSON dumpers)
        # tmp = self.json_dict(**_mod_def_dump_args(kwargs))
        return to_yaml_str(self)

    @classmethod
    def parse_file(cls, path: Union[str, Path]):
        return parse_yaml_file_as(cls, path)

    @classmethod
    def parse_raw(cls, dat: Union[str, bytes], **kwargs):
        try:
            return super().parse_raw(dat, **kwargs)
        except ValidationError:
            return parse_yaml_raw_as(cls, dat)

    def __bytes__(self) -> bytes:
        """Serialize to JSON and return UTF-8 encoded bytes to be written in a file."""
        # add a newline, as otherwise behaviour with text editors will be confusing
        # (e.g. vim automatically adds a trailing newline that it hides)
        # https://stackoverflow.com/questions/729692/why-should-text-files-end-with-a-newline
        return (self.json() + "\n").encode(encoding="utf-8")

    def __str__(self) -> str:
        return self.json(indent=2)



from typing import List, Optional, Union

import pytest
from pydantic import Extra

from metador_core.plugin.metaclass import UndefVersion
from metador_core.plugin.util import register_in_group
from metador_core.schema.core import MetadataSchema, detect_field_overrides
from metador_core.schema.decorators import add_const_fields, override
from metador_core.schema.parser import BaseParser
from metador_core.util.models import updated_fields


def test_schema_metaclass_exceptions():
    with pytest.raises(TypeError):

        class Foo(MetadataSchema):
            ...

        class Bar(MetadataSchema):
            ...

        # only one parent!
        class Qux(Foo, Bar):
            ...

    with pytest.raises(TypeError):
        # forbidden attributes
        class Foo(MetadataSchema):
            __constants__ = {}

    with pytest.raises(TypeError):
        # prevent many config changes
        class Foo(MetadataSchema):
            class Config:
                bla = 5

    with pytest.raises(TypeError):

        @add_const_fields(dict(cnst="my const"))
        class Foo(MetadataSchema):
            ...

        # prevent using names for fields that are parent consts
        class Bar(Foo):
            cnst: str

    with pytest.raises(TypeError):

        class Foo(MetadataSchema):
            class Config:
                extra = Extra.forbid

        # parent extras forbidden, but in child not
        class Bar(Foo):
            class Config:
                extra = Extra.ignore

    with pytest.raises(TypeError):

        class Foo(MetadataSchema):
            class Config:
                extra = Extra.forbid

        # parent extras forbidden, child has new fields
        class Bar(Foo):
            new_field: int


class DummyModel(MetadataSchema):
    Parser = BaseParser

    val: float


class DummyParent(MetadataSchema):
    ovrd: int


@add_const_fields(dict(cnst="my const"))  # adds a const field
@override("ovrd")  # adds an override marker
class Foo(DummyParent):
    class Plugin:  # plugin inner class
        name = "dummy.plugin"
        version = (0, 1, 0)

    ovrd: Optional[str]  # type: ignore
    _priv: str = "hello"

    x: int
    """A number."""

    y: Optional[DummyModel]
    """A schema."""


class Bar(Foo):
    """Test schema."""

    z: bool
    """Some flag."""


def test_schema_metaclass_init():
    # check parent
    assert Foo.Plugin is not None
    assert set(Foo.__constants__.keys()) == {"cnst"}
    assert Foo.__overrides__ == {"ovrd"}

    # check child
    # ----
    # special markers
    assert Bar.Plugin is None

    assert Bar.__overrides__ is not Foo.__overrides__
    assert not Bar.__overrides__

    assert Bar.__constants__ is not Foo.__constants__
    assert Bar.__constants__ == Foo.__constants__

    # (cached) typehints
    assert "z" not in Bar._base_typehints
    assert "x" in Bar._base_typehints
    assert dict(Bar._base_typehints.items()) == dict(Foo._typehints.items())

    assert "z" in Bar._typehints
    assert "x" in Bar._typehints
    assert Bar._typehints["x"] == Foo._typehints["x"]

    # public
    # ----
    # class is pretty-printed
    assert str(Bar).find("Description:") > 0

    # can access partial
    assert DummyModel.Partial.Parser is DummyModel.Parser
    obj = Bar(x=5, z=True)
    prt = Bar.Partial(z=True).merge_with(Bar.Partial(x=5))
    assert prt.from_partial() == obj

    # can access fields
    assert set(iter(Bar.Fields)) == {
        "ovrd",
        "x",
        "y",
        "z",
    }  # no private or const fields
    assert Bar.Fields.x.origin is Foo
    assert Bar.Fields.x.name == "x"
    assert Bar.Fields.x.description is None
    assert Bar.Fields.y.description == "A schema."
    assert Bar.Fields.z.origin is Bar
    assert list(iter(Bar.Fields.x.schemas)) == []
    assert list(iter(Bar.Fields.y.schemas)) == ["DummyModel"]
    assert Bar.Fields.y.schemas.DummyModel is DummyModel


def test_forbidden_types_rejected(schemas_test):
    """Test field hints being checked for forbidden patterns."""

    class InvalidModel(MetadataSchema):
        class Plugin:
            name = "dummy.test"
            version = (0, 1, 0)

        x: Union[int, List[int]]  # <- a forbidden pattern

    with pytest.raises(TypeError):
        register_in_group(schemas_test, InvalidModel, violently=True)

    class InvalidModel2(MetadataSchema):
        Plugin = InvalidModel.Plugin

        x: List[Optional[int]]  # <- forbidden pattern

    with pytest.raises(TypeError):
        register_in_group(schemas_test, InvalidModel2, violently=True)


def test_wrapped_fields_undef_version(schemas_test):
    """Test that UndefVersion infects introspected field schemas."""
    register_in_group(schemas_test, Foo, violently=True)
    GotFoo = schemas_test["dummy.plugin"]

    assert not UndefVersion._is_marked(Foo)
    assert UndefVersion._is_marked(GotFoo)
    assert repr(GotFoo.Fields.y) == repr(Foo.Fields.y)

    GotDummy = GotFoo.Fields.y.schemas.DummyModel
    assert UndefVersion._is_marked(GotDummy)
    assert UndefVersion._unwrap(GotDummy) is DummyModel

    with pytest.raises(TypeError):

        class InvalidSchema(GotDummy):  # <- no well-defined version
            ...

    with pytest.raises(TypeError):

        @register_in_group(schemas_test, violently=True)
        class InvalidSchema2(Bar):
            class Plugin:
                name = "dummy.invalid"
                version = (0, 1, 2)

            fld: GotDummy  # <- no well-defined version


def test_detect_overridden(schemas_test):
    """Cross-check detect_field_overrides with updated_fields."""
    for name in schemas_test.keys():
        schema = schemas_test[name]
        assert detect_field_overrides(schema).issubset(updated_fields(schema))



"""Dummy plugin group and plugins for test cases."""
from metador_core.plugin.interface import PluginBase, PluginGroup
from metador_core.plugin.util import check_implements_method

# ----
# the following are invalid plugingroups to be rejected:


class PGInvalid1:  # not subclass of PluginGroup
    class Plugin:
        name = "invalid"
        version = (0, 1, 0)


class PGInvalid2(PluginGroup):
    class Plugin:
        name = "invalid"
        version = (0, 1, 0)

    # no check_plugin method


class PGInvalid3(PluginGroup):
    # incomplete Plugin class (misses plugin_info_class)
    class Plugin:
        name = "invalid"
        version = "0.1"


class InvalidPGI:
    ...  # invalid plugin_info_class


class PGInvalid4(PluginGroup):
    class Plugin:
        name = "invalid"
        version = (0, 1, 0)
        plugin_info_class = InvalidPGI


# ----
# this is a proper plugingroup that should be accepted and work:


class DummyBase:
    """Base class for dummy plugins."""

    def perform(self):
        """Perform dummy action."""


class DummyPluginInfo(PluginBase):
    """Model for dummy plugin info."""

    something: int


class PGDummy(PluginGroup):
    """Dummy plugin group."""

    class Plugin:
        name = "dummy"
        version = (0, 1, 0)
        plugin_class = DummyBase
        plugin_info_class = DummyPluginInfo

    # must be overridden
    def check_plugin(self, ep_name, plugin):
        return check_implements_method(ep_name, plugin, DummyBase.perform)

    # can be overridden
    def init_plugin(self, plugin):
        return super().init_plugin(plugin)


class DummyPlugin(DummyBase):
    class Plugin:
        name = "test.plugin"
        version = (0, 1, 0)
        something = 1

    def perform(self):
        """Perform dummy action."""


class DummyPluginB(DummyPlugin):
    class Plugin:
        name = "test.plugin"
        version = (0, 1, 1)
        something = 1

    def perform(self):
        ...


class DummyPluginC(DummyPlugin):
    class Plugin:
        name = "test.plugin"
        version = (1, 2, 3)
        something = 1

    def perform(self):
        ...


class DummyPlugin2(DummyPlugin):
    class Plugin:
        name = "test.plugin2"
        version = (0, 1, 0)
        something = 1

    def perform(self):
        ...


# ----
# the following are all invalid dummy plugins to be rejected:


class InvalidPlugin1:
    ...  # no Plugin inner class


class InvalidPlugin2:
    class Plugin:
        name = "abc"  # name mismatch with test.plugin ep
        version = (0, 1, 0)


class InvalidPlugin2b:
    class Plugin:
        name = "test"  # lacks prefix
        version = (0, 1, 0)


class InvalidPlugin3:
    class Plugin:
        name = "test.plugin"
        version = (0, 1, 0)
        # missing "something"


class InvalidPlugin4:  # not subclass of DummyBase
    class Plugin:
        name = "test.plugin"
        version = (0, 1, 0)
        something = 1


class InvalidPlugin5(DummyBase):
    class Plugin:
        name = "test.plugin"
        version = (0, 1, 0)
        something = 1

    # missing "perform" method



import pytest

from metador_core.plugin.util import (
    check_implements_method,
    check_is_subclass,
    is_notebook,
    register_in_group,
)

from .dummy_plugins import PGDummy

# some dummy classes for use in tests:


class A:
    def foo(self):
        pass


class B(A):
    # inherits from A, but does not implement foo
    pass


class C(B):
    # overrides foo that was defined in A
    def foo(self):
        pass


class D:
    # no foo method available
    pass


# ----


def test_check_is_subclass():
    check_is_subclass("B", B, A)
    check_is_subclass("C", C, A)

    with pytest.raises(TypeError):
        check_is_subclass("D", D, A)
    with pytest.raises(TypeError):
        check_is_subclass("A", A, B)


def test_check_implements_method():
    check_implements_method("C", C, A.foo)
    with pytest.raises(TypeError):
        check_implements_method("B", B, A.foo)
    with pytest.raises(TypeError):
        check_implements_method("D", D, A.foo)


# ----


def test_is_notebook():
    # sanity check: tests do not run in a notebook
    assert not is_notebook()


def test_register_in_group(plugingroups_test):
    """Test manual ad-hoc plugin registration."""
    pgs = plugingroups_test

    # should work
    assert "dummy" not in pgs
    register_in_group(pgs, PGDummy, violently=True)
    assert "dummy" in pgs

    # should work as a decorator too
    @register_in_group(pgs, violently=True)
    class PGDummy2(PGDummy):
        class Plugin:
            # must be defined explicitly, or won't pass as plugin
            name = "dummy"
            version = (0, 1, 0)

        def check_plugin(self, ep_name, plugin):
            ...

    # fails:

    with pytest.raises(RuntimeError):
        # not allowed outside notebooks (unless forced)
        register_in_group(pgs, A)

    with pytest.raises(RuntimeError):
        # not suitable (no Plugin inner class)
        register_in_group(pgs, A, violently=True)



"""Metador driver abstraction in order to enable different underlying implementations."""
from enum import Enum
from typing import Any, Optional, Type, Union, cast

import h5py

from ..ih5.container import IH5Record
from .protocols import H5FileLike, OpenMode


class MetadorDriverEnum(Enum):
    """Supported classes that work with MetadorContainer.

    Note that they must be unrelated (i.e. not subclasses of each other).
    """

    HDF5 = h5py.File
    IH5 = IH5Record

    @classmethod
    def to_dict(cls):
        return {x: x.value for x in iter(cls)}


# NOTE: must be duplicated or static checkers can't pick it up
MetadorDriver = Union[h5py.File, IH5Record]
"""Union of all supported classes (for static type check)."""

METADOR_DRIVERS = MetadorDriverEnum.to_dict()
"""Dict representation of MetadorDriverEnum."""

METADOR_DRIVER_CLASSES = tuple(METADOR_DRIVERS.values())
"""Tuple of all supported classes (for instance check)."""

# ----


def get_driver_type(raw_cont: MetadorDriver) -> MetadorDriverEnum:
    """Return the driver type of container (if it is a suitable (sub)class)."""
    for val, cls in METADOR_DRIVERS.items():
        if isinstance(raw_cont, cls):
            return val
    raise ValueError(f"Object not of known container driver type: {raw_cont}")


def get_source(raw_cont: MetadorDriver, driver: MetadorDriverEnum = None) -> Any:
    """Return an object (i.e. input resource(s)) needed to re-open the given container."""
    c = cast(Any, raw_cont)
    driver = driver or get_driver_type(raw_cont)
    if driver == MetadorDriverEnum.HDF5:
        return c.filename
    elif driver == MetadorDriverEnum.IH5:
        return c.ih5_files


def to_h5filelike(
    name_or_obj: Union[MetadorDriver, Any],
    mode: OpenMode = "r",
    *,
    # NOTE: driver takes actual class instead of enum, to also allow subclasses
    driver: Optional[Type[MetadorDriver]] = None,
) -> H5FileLike:
    """Given a container or a resource with a driver, try to return a H5FileLike.

    If first argument is instance of a known driver is returned unchanged.
    Otherwise, will try to open it using the driver (h5py.File by default).

    Returns a H5FileLike compatible with MetadorContainer, or raises ValueError.
    """
    if isinstance(name_or_obj, METADOR_DRIVER_CLASSES):
        # user has passed a h5file-like object already
        return cast(H5FileLike, name_or_obj)
    else:
        # user passed arguments to try instantiating a container object
        driver = driver or h5py.File
        if not issubclass(driver, METADOR_DRIVER_CLASSES):
            msg = f"Passed driver class not supported: {driver}"
            raise ValueError(msg)
        return cast(H5FileLike, driver(cast(Any, name_or_obj), mode))
